# pysam for sam or bam file
import pysam
import time
import copy
from lib.common.os import remove_file_name_suffix, cmd_run, multiprocess_running
from lib.xuyuxing.base.common_command import flag_maker, flag_parse, flag_filter
from lib.common.util import logging_init
from lib.common.genome.seq_base import read_fastq_big, read_fasta_by_faidx
from pysam import VariantFile
from lib.common.math.bin_and_window import split_sequence_to_bins
import pickle
import numpy as np

# for bam or sam file

"""
bam_file = '/lustre/home/xuyuxing/Work/Other/saif/Meth/Con_1.bsmap.bam'

sam file flag:    
1:   read paired
2:   read mapped in proper pair
3:   read unmapped
4:   mate unmapped
5:   read reverse strand
6:   mate reverse strand
7:   first in pair
8:   second in pair
9:   not primary alignment (secondary alignment)
10:  read fails platform/vendor quality checks
11:  read is PCR or optical duplicate
12:  supplementary alignment
"""
# good flag have two filter, need and exclude, used as: flag_filter(83, *good_pair_flag)
good_pair_flag = (flag_maker([1, 2]), flag_maker([3, 4, 10, 12]))
just_pair_flag = (flag_maker([1]), 0)


def parse_bam_by_reads(bam_file):
    """
    parse a bam file, and yield all hit for a read one time
    The bam file should sorted by reads name (samtools sort -n), yield one read one time
    """
    bf = pysam.AlignmentFile(bam_file, 'r')
    read_tmp_record = []
    for r in bf.fetch(until_eof=True):
        if len(read_tmp_record) == 0 or r.query_name == read_tmp_record[-1].query_name:
            read_tmp_record.append(r)
            continue
        else:
            yield read_tmp_record
            read_tmp_record = [r]
    if len(read_tmp_record) != 0:
        yield read_tmp_record
    bf.close()


def pair_read_hit(read_hit_list):
    """
    given a list for hits from same read, get tuple for same template pairs, and return a tuple list
    """

    tmp_id = 0
    read_hit_dir = {}
    for read_hit in read_hit_list:
        read_hit_dir[tmp_id] = read_hit
        tmp_id = tmp_id + 1

    used_hit = []
    output_pair_list = []
    for hit_id in range(0, len(read_hit_list)):
        if hit_id in used_hit:
            continue
        used_hit.append(hit_id)
        hit_tmp = read_hit_dir[hit_id]
        # find good read pair
        if flag_filter(hit_tmp.flag, *good_pair_flag):
            pair_chr = hit_tmp.reference_name
            pair_site = hit_tmp.mpos
            pair_hit_id_list = [j for j in read_hit_dir if
                                (j not in used_hit) and flag_filter(read_hit_dir[j].flag, *good_pair_flag) and
                                read_hit_dir[
                                    j].reference_name == pair_chr and read_hit_dir[j].pos == pair_site]

            if len(pair_hit_id_list) == 0:
                if hit_tmp.is_read1:
                    output_pair_list.append((hit_tmp, None))
                else:
                    output_pair_list.append((None, hit_tmp))
            else:
                pair_hit_id = pair_hit_id_list[0]
                pair_hit_tmp = read_hit_dir[pair_hit_id]
                used_hit.append(pair_hit_id)
                if hit_tmp.is_read1:
                    output_pair_list.append((hit_tmp, pair_hit_tmp))
                else:
                    output_pair_list.append((pair_hit_tmp, hit_tmp))
        elif flag_filter(hit_tmp.flag, *just_pair_flag):
            if hit_tmp.is_read1:
                output_pair_list.append((hit_tmp, None))
            else:
                output_pair_list.append((None, hit_tmp))
        else:
            raise ValueError("Flag should not happen!")

    return output_pair_list


def split_read_sorted_bam(bam_file, reads_number_per_file=10000000, log_file=None):
    module_log = logging_init("split_read_sorted_bam", log_file)
    module_log.info('received a call to "split_read_sorted_bam"')

    with pysam.AlignmentFile(bam_file, "rb") as bf:
        raw_bf_header_dict = bf.header.as_dict()

    output_prefix = remove_file_name_suffix(bam_file, 1) + ".split"

    output_file_list = []

    output_file_id = 0
    output_file_name_tmp = "%s.%d.bam" % (output_prefix, output_file_id)
    module_log.info('begin write %s' % output_file_name_tmp)
    output_file_list.append(output_file_name_tmp)
    tmp_header = raw_bf_header_dict
    tmp_header['CO'] = ['split file %d' % output_file_id]
    bam_out = pysam.AlignmentFile(
        output_file_name_tmp, "wb", header=tmp_header)

    read_num = 0
    start_time = time.time()
    for read_hit_list in parse_bam_by_reads(bam_file):
        read_num = read_num + 1
        for read_hit in read_hit_list:
            bam_out.write(read_hit)

        round_time = time.time()
        if round_time - start_time > 10:
            module_log.info("\tparsed reads: %d" % (read_num))
            start_time = round_time

        if read_num % reads_number_per_file == 0:
            bam_out.close()
            output_file_id = output_file_id + 1
            output_file_name_tmp = "%s.%d.bam" % (
                output_prefix, output_file_id)
            module_log.info('begin write %s' % output_file_name_tmp)
            tmp_header['CO'] = ['split file %d' % output_file_id]
            bam_out = pysam.AlignmentFile(
                output_file_name_tmp, "wb", header=tmp_header)
            output_file_list.append(output_file_name_tmp)

    if not bam_out.is_closed:
        bam_out.close()

    del module_log.handlers[:]
    return output_file_list


def hit_counter(bam_file, log_file=None):
    """
    parse a bam file and return the hit counter for every reads
    The bam file should sorted by read name (samtools sort -n)

    """
    module_log = logging_init("hit_counter", log_file)
    module_log.info('received a call to "hit_counter"')

    read_num = 0
    start_time = time.time()
    hit_count_dist = {}
    for read_hit_list in parse_bam_by_reads(bam_file):
        read_num = read_num + 1

        round_time = time.time()
        if round_time - start_time > 10:
            module_log.info("\tparsed reads: %d" % (read_num))
            start_time = round_time

        read_hit_list = pair_read_hit(read_hit_list)
        good_pair_hit_num = len(
            [i for i in read_hit_list if
             (not i[0] is None) and (not i[1] is None) and flag_filter(i[0].flag, *good_pair_flag) and flag_filter(
                 i[1].flag, *good_pair_flag)])

        if good_pair_hit_num not in hit_count_dist:
            hit_count_dist[good_pair_hit_num] = 0

        hit_count_dist[good_pair_hit_num] = hit_count_dist[good_pair_hit_num] + 1

    module_log.info('finished parser"')
    del module_log.handlers[:]
    return hit_count_dist


def split_sorted_bam_by_contig(bam_file, log_file=None, change_header=True):
    module_log = logging_init("split_sorted_bam_by_contig", log_file)
    module_log.info('received a call to "split_sorted_bam_by_contig"')

    output_prefix = remove_file_name_suffix(bam_file, 1)

    with pysam.AlignmentFile(bam_file, "rb") as bf:
        raw_bf_header_dict = bf.header.as_dict()

    contig_list = [i['SN'] for i in raw_bf_header_dict['SQ']]
    contig_file_dict = {}

    for i in raw_bf_header_dict['SQ']:
        contig = i['SN']
        contig_length = i['LN']
        contig_file_dict[contig] = {}
        contig_file_dict[contig]['file'] = output_prefix + \
            "." + contig + ".bam"
        contig_file_dict[contig]['head'] = {}
        for j in raw_bf_header_dict:
            if j == 'SQ':
                contig_file_dict[contig]['head']['SQ'] = [
                    {'SN': contig, 'LN': contig_length}]
            else:
                contig_file_dict[contig]['head'][j] = raw_bf_header_dict[j]

    bf = pysam.AlignmentFile(bam_file, 'r')
    for contig in contig_list:
        output_file = contig_file_dict[contig]['file']
        module_log.info('begin write %s' % output_file)
        tmp_header = contig_file_dict[contig]['head']
        tmp_header['CO'] = ['split file %s' % output_file]
        # print(tmp_header)
        if change_header:
            bam_out = pysam.AlignmentFile(output_file, "wb", header=tmp_header)
        else:
            bam_out = pysam.AlignmentFile(
                output_file, "wb", header=raw_bf_header_dict)

        for r in bf.fetch(contig):
            bam_out.write(r)

        bam_out.close()

    bf.close()

    del module_log.handlers[:]
    return contig_file_dict


def get_read_name_list_from_fastq(fastq_file_name, output_file_name):
    """
    get_read_name_list_from_fastq
    """
    with open(output_file_name, 'w') as f:
        for record in read_fastq_big(fastq_file_name, gzip_flag=True):
            f.write(record[0].seqname_short() + "\n")

    return output_file_name


def index_bam_file(bam_file):
    """
    samtools index bam_file
    """
    bamfile = pysam.AlignmentFile(bam_file, 'rb')
    name_indexed = pysam.IndexedReads(bamfile)
    name_indexed.build()
    return name_indexed


def extract_aln_by_reads(reads_name, name_indexed):
    output_list = []
    try:
        name_indexed.find(reads_name)
    except KeyError:
        pass
    else:
        iterator = name_indexed.find(reads_name)
        output_list = []
        for x in iterator:
            output_list.append(x)

    return output_list


def get_best_alignment_match_number(read_aln_list):
    valid_aln_list = [alignedsegment for alignedsegment in read_aln_list if not (
        alignedsegment.is_secondary or alignedsegment.is_supplementary)]

    # TODO


def extract_reads_from_bam_file(reads_name_list, input_bam_file, output_bam_file):

    bamfile = pysam.AlignmentFile(input_bam_file, 'rb')
    name_indexed = pysam.IndexedReads(bamfile)
    name_indexed.build()
    header = bamfile.header.copy()
    out = pysam.Samfile(output_bam_file, 'wb', header=header)

    for name in reads_name_list:
        try:
            name_indexed.find(name)
        except KeyError:
            pass
        else:
            iterator = name_indexed.find(name)
            for x in iterator:
                out.write(x)


# for bcf or vcf file


def compare_two_bcf_file(A_bcf_file, B_bcf_file):

    A_bcf = VariantFile(A_bcf_file)
    A_site_dict = {}
    for rec in A_bcf.fetch():
        A_site_dict[(rec.contig, rec.pos)] = rec

    B_bcf = VariantFile(B_bcf_file)
    B_site_dict = {}
    for rec in B_bcf.fetch():
        B_site_dict[(rec.contig, rec.pos)] = rec

    onlyA_site = set(A_site_dict.keys()) - set(B_site_dict.keys())
    onlyB_site = set(B_site_dict.keys()) - set(A_site_dict.keys())
    AB_site = set(A_site_dict.keys()) & set(B_site_dict.keys())

    return list(onlyA_site), list(onlyB_site), list(AB_site), A_site_dict, B_site_dict


def filter_EMS_snp(rec):
    return rec.alleles == ('G', 'A') or rec.alleles == ('C', 'T')


def get_AD_dict(vcf_line):
    vs = vcf_line.strip()

    v_info = vs.split("\t")
    g_type = [v_info[3]] + v_info[4].split(",")

    info_dict = dict(zip(v_info[8].split(":"), v_info[9].split(":")))

    if 'AD' in info_dict:
        AD_tuple = [int(i) for i in info_dict['AD'].split(",")]
        output = {i: j for i, j in zip(g_type, AD_tuple) if i != '<NON_REF>'}
        output = {i:output[i] for i in output if output[i] != 0}
        return output
    else:
        if 'DP' in info_dict:
            DP_value = int(info_dict['DP'])
            return {g_type[0]: DP_value}
        else:
            return None

def get_AD_dict_more_sample(vcf_line, sample_num=1):
    vs = vcf_line.strip()

    v_info = vs.split("\t")
    g_type = [v_info[3]] + v_info[4].split(",")

    output_list = []
    for rank in range(1,sample_num+1):
        info_dict = dict(zip(v_info[8].split(":"), v_info[8+rank].split(":")))

        if 'AD' in info_dict:
            AD_tuple = [int(i) for i in info_dict['AD'].split(",")]
            output = {i: j for i, j in zip(g_type, AD_tuple) if i != '<NON_REF>'}
            output = {i:output[i] for i in output if output[i] != 0}
        else:
            if 'DP' in info_dict:
                DP_value = int(info_dict['DP'])
                output = {g_type[0]: DP_value}

        output_list.append(output)

    return output_list


def BSA_eu_dist(g1_AD_dict, g2_AD_dict, min_depth=8, max_depth=250):
    c = ['A', 'T', 'C', 'G']

    sum1 = sum([g1_AD_dict[i] for i in c if i in g1_AD_dict])
    sum2 = sum([g2_AD_dict[i] for i in c if i in g2_AD_dict])

    if sum1 < min_depth or sum2 < min_depth or sum1 > max_depth or sum2 > max_depth:
        return 0.0

    g1 = {}
    for i in c:
        if sum1 == 0 or i not in g1_AD_dict:
            g1[i] = 0.0
        else:
            g1[i] = g1_AD_dict[i]/sum1

    g2 = {}
    for i in c:
        if sum2 == 0 or i not in g2_AD_dict:
            g2[i] = 0.0
        else:
            g2[i] = g2_AD_dict[i]/sum2

    return(np.sqrt(sum([np.square(g1[i] - g2[i]) for i in c])))



def BSA_ED_in_one_range(chr_id, start, end, group1_g_bcf, group2_g_bcf, min_depth=8, max_depth=250):
    g1_bcf = VariantFile(group1_g_bcf)
    g2_bcf = VariantFile(group2_g_bcf)

    output_dict = {}
    output_dict[chr_id] = {}

    g1s_dict = {i.pos:i for i in g1_bcf.fetch(chr_id, start, end)}
    g2s_dict = {i.pos:i for i in g2_bcf.fetch(chr_id, start, end)}

    for i in range(start, end + 1):
        site = i + 1

        if site in g1s_dict and site in g2s_dict:


            g1s = g1s_dict[site]
            g2s = g2s_dict[site]

            g1_AD_dict = get_AD_dict(g1s.__str__())
            g2_AD_dict = get_AD_dict(g2s.__str__())

            if g1_AD_dict is None or g2_AD_dict is None:
                continue

            ED = BSA_eu_dist(g1_AD_dict, g2_AD_dict, min_depth, max_depth)

            if ED > 0.0:
                output_dict[chr_id][site] = ED

    return output_dict

def BSA_ED_filter(group1_g_bcf, group2_g_bcf, genome_file, min_depth=8, max_depth=250, bin_length=1000000, threads=10):
    """
    # bcf should have index
    bcftools view -O b -o SGJ-R.g.bcf SGJ-R.g.vcf
    bcftools index SGJ-R.g.bcf
    """
    fasta_dict = read_fasta_by_faidx(genome_file)

    args_list = []
    output_dict = {}
    for chr_id in fasta_dict:
        output_dict[chr_id] = {}
        length = fasta_dict[chr_id].len()
        for i, s, e in split_sequence_to_bins(length, bin_length):
            args_list.append((chr_id, s, e, group1_g_bcf, group2_g_bcf, min_depth, max_depth))

    mp_out_dict = multiprocess_running(
        BSA_ED_in_one_range, args_list, threads, silence=False)

    for i in mp_out_dict:
        tmp_dict = mp_out_dict[i]['output']
        for j in tmp_dict:
            for k in tmp_dict[j]:
                output_dict[j][k] = tmp_dict[j][k]

    return output_dict

def BSA_mutmap_in_one_range(chr_id, start, end, mutmap_vcf, min_depth=8, max_depth=250):
    bcf = VariantFile(mutmap_vcf)

    output_dict = {}
    output_dict[chr_id] = {}

    gs_dict = {i.pos:i for i in bcf.fetch(chr_id, start, end)}

    for i in range(start, end + 1):
        site = i + 1

        if site in gs_dict:

            gs = gs_dict[site]

            g1_AD_dict, g2_AD_dict = get_AD_dict_more_sample(gs.__str__(), sample_num=2)

            if g1_AD_dict == {} or g2_AD_dict == {}:
                continue

            ED = BSA_eu_dist(g1_AD_dict, g2_AD_dict, min_depth, max_depth)

            # print(chr_id, site, g1_AD_dict, g2_AD_dict, ED)

            if ED > 0.0:
                output_dict[chr_id][site] = ED

    return output_dict

def BSA_mutmap_filter(mutmap_vcf, genome_file, min_depth=8, max_depth=250, bin_length=1000000, threads=10):
    """
    # bcf should have index
    bcftools view -O b -o SGJ-R.g.bcf SGJ-R.g.vcf
    bcftools index SGJ-R.g.bcf
    """
    fasta_dict = read_fasta_by_faidx(genome_file)

    args_list = []
    output_dict = {}
    for chr_id in fasta_dict:
        output_dict[chr_id] = {}
        length = fasta_dict[chr_id].len()
        for i, s, e in split_sequence_to_bins(length, bin_length):
            args_list.append((chr_id, s, e, mutmap_vcf, min_depth, max_depth))

    mp_out_dict = multiprocess_running(
        BSA_mutmap_in_one_range, args_list, threads, silence=False)

    for i in mp_out_dict:
        tmp_dict = mp_out_dict[i]['output']
        for j in tmp_dict:
            for k in tmp_dict[j]:
                output_dict[j][k] = tmp_dict[j][k]

    return output_dict


if __name__ == "__main__":

    for i in ["ZJXA5242", "ZJXA5243", "ZJXA5244", "ZJXA5245", "ZJXA5246"]:
        A_bcf_file = '/lustre/home/xuyuxing/Database/Cuscuta/Cau/raw_data/mutant_reseq2_202010/SURVEY.vcf.gz'
        B_bcf_file = '/lustre/home/xuyuxing/Database/Cuscuta/Cau/raw_data/mutant_reseq2_202010/%s.snp.filter.vcf.gz' % i
        output_B_file = '/lustre/home/xuyuxing/Database/Cuscuta/Cau/raw_data/mutant_reseq2_202010/%s.snp.filter.again.vcf' % i

        onlyA_site, onlyB_site, AB_site, A_site_dict, B_site_dict = compare_two_bcf_file(
            A_bcf_file, B_bcf_file)

        bcf_in = VariantFile(B_bcf_file)  # auto-detect input format
        bcf_out = VariantFile(output_B_file, 'w', header=bcf_in.header)
        num = 0
        for i in onlyB_site:
            rec = B_site_dict[i]
            print(rec.alleles)
            if filter_EMS_snp(rec):
                print('True')
                num += 1
                bcf_out.write(rec)
        bcf_out.close()

    for i in ["ZJXA5242", "ZJXA5243", "ZJXA5244", "ZJXA5245", "ZJXA5246"]:
        good_vcf_file = '/lustre/home/xuyuxing/Database/Cuscuta/Cau/raw_data/mutant_reseq2_202010/%s.snp.filter.again.vcf' % i
        output_file = '/lustre/home/xuyuxing/Database/Cuscuta/Cau/raw_data/mutant_reseq2_202010/%s.snp.top30.txt' % i

        B_bcf = VariantFile(good_vcf_file)
        B_site_dict = {}
        for rec in B_bcf.fetch():
            B_site_dict[(rec.contig, rec.pos)] = rec

        with open(output_file, 'w') as f:
            for id_now in sorted(B_site_dict.keys(), key=lambda x: B_site_dict[x].qual, reverse=True)[0:30]:
                chr_id, site = id_now
                rec = B_site_dict[id_now]

                f.write(rec.__str__())

                cmd_string = "samtools faidx /lustre/home/xuyuxing/Database/Cuscuta/Cau/raw_data/mutant_reseq2_202010/Cuscuta.genome.v1.1.fasta %s:%d-%d" % (
                    chr_id, site-200, site+200)
                fl, o, e = cmd_run(cmd_string)
                f.write(o+"\n")

    # BSA
    import os
    import pickle

    group1_g_bcf = '/lustre/home/xuyuxing/Work/Other/Guojin/BSA/tomato_def1/SGJ-R.g.bcf'
    group2_g_bcf = '/lustre/home/xuyuxing/Work/Other/Guojin/BSA/tomato_def1/SGJ-S.g.bcf'
    genome_file = '/lustre/home/xuyuxing/Work/Other/Guojin/tomato_collinear/genome_data/Sly/S_lycopersicum_chromosomes.4.00.fa'
    
    ED_dict_pickle = '/lustre/home/xuyuxing/Work/Other/Guojin/BSA/tomato_def1/BSA_ED.pyb'

    if not os.path.exists(ED_dict_pickle):

        ED_dict = BSA_ED_filter(group1_g_bcf, group2_g_bcf, genome_file, min_depth=8, max_depth=250, bin_length=1000000, threads=56)

        OUT = open(ED_dict_pickle, 'wb')
        pickle.dump(ED_dict, OUT)
        OUT.close()
    
    else:
        TEMP = open(ED_dict_pickle, 'rb')
        ED_dict = pickle.load(TEMP)
        TEMP.close()

    g1_bcf = VariantFile(group1_g_bcf)
    g2_bcf = VariantFile(group2_g_bcf)


    def ems_mutant(WT_like_AD_dict, mutant_like_AD_dict, A='WT', more_strict=False):
        True_flag = False
        strict_flag = False

        if A == 'WT':
            # WT_like_AD_dict vs mutant_like_AD_dict: G/A, GA/A, C/T, CT/T
            if len(mutant_like_AD_dict) == 1:
                if 'A' in mutant_like_AD_dict:
                    if len(WT_like_AD_dict) == 1 and 'G' in WT_like_AD_dict:
                        True_flag = True

                    if len(WT_like_AD_dict) == 2 and 'G' in WT_like_AD_dict and 'A' in WT_like_AD_dict:
                        True_flag = True
                        if WT_like_AD_dict['G'] > WT_like_AD_dict['A']:
                            strict_flag = True

                elif 'T' in mutant_like_AD_dict:
                    if len(WT_like_AD_dict) == 1 and 'C' in WT_like_AD_dict:
                        True_flag = True

                    if len(WT_like_AD_dict) == 2 and 'C' in WT_like_AD_dict and 'T' in WT_like_AD_dict:
                        True_flag = True
                        if WT_like_AD_dict['C'] > WT_like_AD_dict['T']:
                            strict_flag = True

        elif A == 'MT':
            # WT_like_AD_dict vs mutant_like_AD_dict: G/A, G/GA, C/T, C/CT
            if len(WT_like_AD_dict) == 1:
                if 'G' in WT_like_AD_dict:
                    if len(mutant_like_AD_dict) == 1 and 'A' in mutant_like_AD_dict:
                        True_flag = True

                    if len(mutant_like_AD_dict) == 2 and 'G' in mutant_like_AD_dict and 'A' in mutant_like_AD_dict:
                        True_flag = True
                        if mutant_like_AD_dict['A'] > mutant_like_AD_dict['G']:
                            strict_flag = True

                elif 'C' in WT_like_AD_dict:
                    if len(mutant_like_AD_dict) == 1 and 'T' in mutant_like_AD_dict:
                        True_flag = True

                    if len(mutant_like_AD_dict) == 2 and 'C' in mutant_like_AD_dict and 'T' in mutant_like_AD_dict:
                        True_flag = True
                        if mutant_like_AD_dict['T'] > mutant_like_AD_dict['C']:
                            strict_flag = True

        if more_strict:
            return strict_flag
        else:
            return True_flag


    from lib.xuyuxing.math.stats import fisher_enrichment

    def heterozygote_pvalue(AD_dict, ratio_dict):
        """
        ratio_dict = {'G':2,'A':1}
        """
        
        for i in ratio_dict:
            if i not in AD_dict:
                return None
        
        key_list = list(ratio_dict.keys())

        sum_obs = sum([AD_dict[i] for i in key_list])
        sum_ratio = sum([ratio_dict[i] for i in key_list])

        Contingency_table = [[(ratio_dict[i] / sum_ratio) * sum_obs for i in key_list], [AD_dict[i] for i in key_list]]

        # print(Contingency_table)

        oddsratio, pvalue = fisher_enrichment(Contingency_table)

        return pvalue

    num = 0 
    bsa_output_dict = {}
    for chr_id in ED_dict:
        for site in ED_dict[chr_id]:
            if ED_dict[chr_id][site] > 0.8:

                g1s_dict = {i.pos:i for i in g1_bcf.fetch(chr_id, site-100, site+100)}
                g2s_dict = {i.pos:i for i in g2_bcf.fetch(chr_id, site-100, site+100)}

                g1s = g1s_dict[site]
                g2s = g2s_dict[site]

                g1_AD_dict = get_AD_dict(g1s.__str__())
                g2_AD_dict = get_AD_dict(g2s.__str__())

                if ems_mutant(g1_AD_dict, g2_AD_dict, A='WT', more_strict=True):
                    num += 1

                    ratio_dict1 = {'G':2,'A':1}
                    ratio_dict2 = {'C':2,'T':1}

                    pvalue = heterozygote_pvalue(g1_AD_dict, ratio_dict1) or heterozygote_pvalue(g1_AD_dict, ratio_dict2)

                    bsa_output_dict[(chr_id,site)] = ED_dict[chr_id][site],pvalue,g1_AD_dict,g2_AD_dict

    gff_file = '/lustre/home/xuyuxing/Work/Other/Guojin/tomato_collinear/genome_data/Sly/ITAG4.1_gene_models.gff'

    from lib.common.genome.genome_feature2 import read_gff_file

    gf_dict = read_gff_file(gff_file)
    gene_dict = gf_dict['gene']

    from interlap import InterLap
    cds_interlap = {}
    for g_id in gene_dict:
        gene = gene_dict[g_id]
        chr_id = gene.chr_id
        if chr_id not in cds_interlap:
            cds_interlap[chr_id] = {
                "+":InterLap(),
                "-":InterLap(),
            }

        for mRNA in gene.sub_features:
            for cds in mRNA.sub_features:
                if cds.type == 'CDS':
                    cds_interlap[chr_id][cds.strand].add((cds.start, cds.end, cds))

    gene_interlap = {}
    for g_id in gene_dict:
        gene = gene_dict[g_id]
        chr_id = gene.chr_id
        if chr_id not in gene_interlap:
            gene_interlap[chr_id] = {
                "+":InterLap(),
                "-":InterLap(),
            }

        gene_interlap[chr_id][gene.strand].add((gene.start, gene.end, gene))

    bas_cds_dict = {}
    for chr_id, site in bsa_output_dict:
        over_cds_list = list(cds_interlap[chr_id]['+'].find((site, site))) + list(cds_interlap[chr_id]['-'].find((site, site)))
        if len(over_cds_list) > 0:
            print(chr_id, site, bsa_output_dict[(chr_id, site)])
            bas_cds_dict[(chr_id, site)] = over_cds_list

    bas_gene_dict = {}
    for chr_id, site in bsa_output_dict:
        over_gene_list = list(gene_interlap[chr_id]['+'].find((site, site))) + list(gene_interlap[chr_id]['-'].find((site, site)))
        if len(over_gene_list) > 0:
            print(chr_id, site, bsa_output_dict[(chr_id, site)], over_gene_list[0][2].id)
            bas_gene_dict[(chr_id, site)] = over_gene_list


    cds = bas_cds_dict[('SL4.0ch10',28051333)][0][2]
    vars(cds)

    cds = bas_cds_dict[('SL4.0ch00', 166925)][0][2]
    vars(cds)                

    cds = bas_cds_dict[('SL4.0ch07', 8202167)][0][2]
    vars(cds)          

    with open("/lustre/home/xuyuxing/Work/Other/Guojin/BSA/tomato_def1/all.site.txt", 'w') as f:
        for i in bsa_output_dict:
            a=bsa_output_dict[i]
            f.write("%s\t%s\t%s\t%s\t%s, %s\n" % (i[0],i[1],a[0],a[1],a[2].__str__(),a[3].__str__()))

    # BSA2
    import os
    import pickle

    mutmap_vcf = '/lustre/home/xuyuxing/Work/Other/Guojin/BSA/tomato_def1/mutmap_dir/30_vcf/mutmap.vcf.gz'
    genome_file = '/lustre/home/xuyuxing/Work/Other/Guojin/tomato_collinear/genome_data/Sly/S_lycopersicum_chromosomes.4.00.fa'
    
    ED_dict_pickle = '/lustre/home/xuyuxing/Work/Other/Guojin/BSA/tomato_def1/mutmap_dir/BSA_ED.pyb'

    if not os.path.exists(ED_dict_pickle):
        ED_dict = BSA_mutmap_filter(mutmap_vcf, genome_file, min_depth=8, max_depth=250, bin_length=1000000, threads=56)

        OUT = open(ED_dict_pickle, 'wb')
        pickle.dump(ED_dict, OUT)
        OUT.close()
    
    else:
        TEMP = open(ED_dict_pickle, 'rb')
        ED_dict = pickle.load(TEMP)
        TEMP.close()

    mm_bcf = VariantFile(mutmap_vcf)

    num = 0 
    bsa_output_dict = {}
    for chr_id in ED_dict:
        for site in ED_dict[chr_id]:
            if ED_dict[chr_id][site] > 0.5:
                

                mm_dict = {i.pos:i for i in mm_bcf.fetch(chr_id, site-100, site+100)}
                
                gs = mm_dict[site]

                g1_AD_dict, g2_AD_dict = get_AD_dict_more_sample(gs.__str__(), sample_num=2)

                if ems_mutant(g1_AD_dict, g2_AD_dict, A='WT', more_strict=True):
                    num += 1

                    ratio_dict1 = {'G':2,'A':1}
                    ratio_dict2 = {'C':2,'T':1}

                    pvalue = heterozygote_pvalue(g1_AD_dict, ratio_dict1) or heterozygote_pvalue(g1_AD_dict, ratio_dict2)

                    bsa_output_dict[(chr_id,site)] = ED_dict[chr_id][site],pvalue,g1_AD_dict,g2_AD_dict

    gff_file = '/lustre/home/xuyuxing/Work/Other/Guojin/tomato_collinear/genome_data/Sly/ITAG4.1_gene_models.gff'

    from lib.common.genome.genome_feature2 import read_gff_file

    gf_dict = read_gff_file(gff_file)
    gene_dict = gf_dict['gene']

    from interlap import InterLap
    cds_interlap = {}
    for g_id in gene_dict:
        gene = gene_dict[g_id]
        chr_id = gene.chr_id
        if chr_id not in cds_interlap:
            cds_interlap[chr_id] = {
                "+":InterLap(),
                "-":InterLap(),
            }

        for mRNA in gene.sub_features:
            for cds in mRNA.sub_features:
                if cds.type == 'CDS':
                    cds_interlap[chr_id][cds.strand].add((cds.start, cds.end, cds))

    gene_interlap = {}
    for g_id in gene_dict:
        gene = gene_dict[g_id]
        chr_id = gene.chr_id
        if chr_id not in gene_interlap:
            gene_interlap[chr_id] = {
                "+":InterLap(),
                "-":InterLap(),
            }

        gene_interlap[chr_id][gene.strand].add((gene.start, gene.end, gene))

    bas_cds_dict = {}
    for chr_id, site in bsa_output_dict:
        over_cds_list = list(cds_interlap[chr_id]['+'].find((site, site))) + list(cds_interlap[chr_id]['-'].find((site, site)))
        if len(over_cds_list) > 0:
            print(chr_id, site, bsa_output_dict[(chr_id, site)])
            bas_cds_dict[(chr_id, site)] = over_cds_list

    bas_gene_dict = {}
    for chr_id, site in bsa_output_dict:
        over_gene_list = list(gene_interlap[chr_id]['+'].find((site, site))) + list(gene_interlap[chr_id]['-'].find((site, site)))
        if len(over_gene_list) > 0:
            print(chr_id, site, bsa_output_dict[(chr_id, site)], over_gene_list[0][2].id)
            bas_gene_dict[(chr_id, site)] = over_gene_list


    cds = bas_cds_dict[('SL4.0ch10',28051333)][0][2]
    vars(cds)

    cds = bas_cds_dict[('SL4.0ch00', 166925)][0][2]
    vars(cds)                

    cds = bas_cds_dict[('SL4.0ch07', 8202167)][0][2]
    vars(cds)          

    with open("/lustre/home/xuyuxing/Work/Other/Guojin/BSA/tomato_def1/all.site.txt", 'w') as f:
        for i in bsa_output_dict:
            a=bsa_output_dict[i]
            f.write("%s\t%s\t%s\t%s\t%s, %s\n" % (i[0],i[1],a[0],a[1],a[2].__str__(),a[3].__str__()))


import re
import gzip
import numpy as np
import skbio
import os
import time
import random
import math
from Bio import SeqIO
from Bio import Seq
from collections import OrderedDict
from pyfaidx import Fasta
from lib.common.util import logging_init
import lib.common.sqlite_command as sc
from lib.common.os import remove_file_name_suffix, cmd_run
# from lib.common.genome.genome_feature2 import cds_judgment


# tools for seq as string

def slice_scaffold_by_N(sequence, scaffold_maker="N"):
    """
    some scaffold sequence have many N in the seq, I can find the start and end for each Non-N seq, e.g. contig. I will
    return a list which have tuple(start,end) 0-base
    """
    subseqs = sequence.split(scaffold_maker)
    sub_seq_slice = []
    pointer = 0
    for i in subseqs:
        if len(i) == 0:
            pointer = pointer + 1
            continue
        sub_seq_slice.append((pointer, pointer + len(i)))
        pointer = pointer + len(i) + 1
    return sub_seq_slice


def contig_cutting(sequence, step, length, pointer_start=1):
    """
    cutting a sequence as windows, you can give me step and window length. I will return a generator which every one is
    start point and window sequence. pointer_start is the begin pointer it just for ID, will not change sequence cutting
    """
    seq_len = len(sequence)
    sub_seq_dict = {}
    for i in range(pointer_start - 1, seq_len, step):
        if seq_len - i > length:
            sub_seq = sequence[i:i + length]
            yield (pointer_start, sub_seq)
            sub_seq_dict[pointer_start] = sub_seq
        else:
            sub_seq = sequence[i:seq_len]
            yield (pointer_start, sub_seq)
            break
        pointer_start = pointer_start + step


def scaffold_cutting(sequence, step=500, length=1000, Consider_scaffold=True, shift_start=0):
    """
    cutting a sequence as windows, you can give me step and window length. I will return a generator which every one is
    start point and window sequence. pointer_start is the begin pointer it just for ID, will not change sequence cutting
    If the sequence is scaffold, I can split scaffold by N firstly, and get contig to make windows, the pointer will follow
    scaffold which based on 1
    """
    if Consider_scaffold is True:
        sub_seq_slice = slice_scaffold_by_N(sequence)
        # output_dict = {}
        for start, end in sub_seq_slice:
            sub_seq = sequence[start:end]
            for pointer, sub_sub_seq in contig_cutting(sub_seq, step, length, start + 1):
                yield (pointer, sub_sub_seq)
                # output_dict[pointer] = sub_sub_seq
    else:
        if shift_start != 0:
            # print(shift_start)
            yield (1, sequence[1 - 1:shift_start])

        # output_dict = {}
        for pointer, sub_seq in contig_cutting(sequence, step, length, shift_start + 1):
            yield (pointer, sub_seq)
            # output_dict[pointer] = sub_seq

    # return output_dict


def reverse_complement(seq, RNA=False):
    """
    #>>> seq = "TCGGinsGCCC"
    #>>> print "Reverse Complement:"
    #>>> print(reverse_complement(seq))
    #GGGCinsCCGA
    """
    alt_map = {'ins': '0'}
    if RNA:
        complement = {'A': 'U', 'C': 'G', 'G': 'C', 'U': 'A'}
    else:
        complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}
    # for k, v in alt_map.iteritems():
    for k, v in alt_map.items():
        seq = seq.replace(k, v)
    bases = list(seq)
    bases = reversed([complement.get(base, base) for base in bases])
    bases = ''.join(bases)
    # for k, v in alt_map.iteritems():
    for k, v in alt_map.items():
        bases = bases.replace(v, k)
    return bases


def reverse_seq(seq):
    """just reverse"""
    return "".join(reversed(seq))


def complement_seq(seq):
    """just complement"""
    if 'U' in seq:
        complement = {'A': 'U', 'C': 'G', 'G': 'C', 'U': 'A'}
    else:
        complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}
    bases = list(seq)
    bases = [complement.get(base, base) for base in bases]
    bases = ''.join(bases)
    return bases


def sub_seq(seq, site_start, site_end, strand, RNA):
    """get sub set of a seq"""
    start = min(site_start, site_end)
    end = max(site_start, site_end)
    sub_seq_out = seq[start - 1:end]

    if RNA is True:
        sub_seq_out = sub_seq_out.replace("T", "U")
    if strand == "-":
        sub_seq_out = reverse_complement(sub_seq_out, RNA)

    return sub_seq_out


def sequence_kmer_frequencies(sequence_input, kmer):
    s = skbio.Sequence(sequence_input)
    freqs = s.kmer_frequencies(kmer, overlap=False)
    return freqs


def sequence_entropy(sequence_input, kmer):
    freqs = sequence_kmer_frequencies(sequence_input, kmer)
    t_num = sum([freqs[i] for i in freqs])

    entropy = 0
    for i in freqs:
        p = freqs[i]/t_num
        entropy = entropy - (p * math.log(p, 2))

    return entropy


def random_sequence(seq_char, length, random_seed=None):
    """
    get random sequence
    :param seq_char: for DNA, you can give "ATCG"
    """
    random.seed(random_seed)
    sa = []
    for i in range(length):
        sa.append(random.choice(seq_char))
    salt = ''.join(sa)
    return salt


def get_seq_index_ignore_gap(give_site, seq_with_gap, seq_start=1, gap_chr='-'):
    """
    seq_with_gap = "------------ATGC---CAGTCA---ACG-GCATGCTA"
    give_site = 5
    seq_start = 1
    gap_chr = '-'

    give a seq with gap, and a site number, return index in gap seq
    """
    go_on_flag = True
    round_use_range = (0, 0)
    add_length = give_site - seq_start + 1
    round_true_site = 0
    while (go_on_flag):
        round_use_range = (
            round_use_range[1] + 1, round_use_range[1] + add_length)
        round_seq = seq_with_gap[round_use_range[0] - 1:round_use_range[1]]
        round_true_site = round_true_site + \
            len(round_seq) - round_seq.count(gap_chr)
        add_length = round_seq.count(gap_chr)
        print(round_use_range, round_seq, round_true_site, add_length)

        if add_length == 0:
            go_on_flag = False

    return round_use_range[1]


# tools for fasta file

class BioSeq(object):
    def __init__(self, seq, seqname=None, seq_type='nucl'):
        self.seqname = seqname
        self.seq = seq
        self.seq_type = seq_type

    nucl = ["A", "T", "C", "G", "U"]
    nucl_degenerate = ["M", "S", "W", "B", "D", "R", "H", "Y", "V", "K"]
    any_nucl = ["N"]
    prot = ["A", "P", "B", "Q", "C", "R", "D", "S", "E", "T", "F", "U", "G", "V", "H", "W", "I", "Y", "K", "Z", "L",
            "M", "N"]
    any_prot = ["X"]
    stop_site = ["*"]
    gap_site = ['-']

    def __str__(self):
        return self.seq

    def seqs_length(self):
        return len(self.seq)

    def seq_clean(self, degenerate_sites_allowed=False, gap_allowed=False,
                  translation_stop_allowed=False, replace_flag=True):
        # make code
        if self.seq_type == 'nucl':
            seq_type_flag = 0b0001
        elif self.seq_type == 'prot':
            seq_type_flag = 0b0000

        dsa_flag = degenerate_sites_allowed
        if dsa_flag:
            dsa_flag = 0b0010

        ga_flag = gap_allowed
        if ga_flag:
            ga_flag = 0b0100

        tsa_flag = translation_stop_allowed
        if tsa_flag:
            tsa_flag = 0b1000

        bit_field = seq_type_flag + dsa_flag + ga_flag + tsa_flag

        if bit_field == 0b0111:
            all_allowed = self.nucl + self.any_nucl + self.nucl_degenerate + self.gap_site
            safe_remove = []
            safe_replace = self.any_nucl
        elif bit_field == 0b0011:
            all_allowed = self.nucl + self.any_nucl + self.nucl_degenerate
            safe_remove = self.gap_site
            safe_replace = self.any_nucl
        elif bit_field == 0b0101:
            all_allowed = self.nucl + self.any_nucl + self.gap_site
            safe_remove = []
            safe_replace = self.any_nucl
        elif bit_field == 0b0001:
            all_allowed = self.nucl + self.any_nucl
            safe_remove = self.gap_site
            safe_replace = self.any_nucl
        elif bit_field == 0b0000:
            all_allowed = self.prot + self.any_prot
            safe_remove = self.gap_site + self.stop_site
            safe_replace = self.any_prot
        elif bit_field == 0b0100:
            all_allowed = self.prot + self.any_prot + self.gap_site
            safe_remove = self.stop_site
            safe_replace = self.any_prot
        elif bit_field == 0b1100:
            all_allowed = self.prot + self.any_prot + self.gap_site + self.stop_site
            safe_remove = []
            safe_replace = self.any_prot
        elif bit_field == 0b1000:
            all_allowed = self.prot + self.any_prot + self.stop_site
            safe_remove = self.gap_site
            safe_replace = self.any_prot
        else:
            raise ValueError("Bad input flag")

        new_seq = self.seq
        char_site = set(new_seq)
        bad_flag = False
        for i in char_site:
            if i in all_allowed:
                continue
            elif i in safe_remove:
                if i == "*":
                    new_seq = re.sub("\*", "", new_seq)
                else:
                    new_seq = re.sub(i, "", new_seq)
            else:
                if replace_flag:
                    if i == "*":
                        new_seq = re.sub("\*", "", new_seq)
                    else:
                        new_seq = re.sub(i, safe_replace[0], new_seq)
                bad_flag = True

        if replace_flag:
            self.seq = new_seq
        else:
            if bad_flag:
                return False
            else:
                return True

    def seqs_upper(self):
        self.seq = self.seq.upper()

    def seqs_lower(self):
        self.seq = self.seq.lower()

    def no_wrap(self):
        self.seq = re.sub("\n", "", self.seq)

    def wrap(self, line_length=75):
        raw_seq = str(self.seq)
        raw_seq = re.sub("\n", "", raw_seq)
        new_seq = ""

        for j in (raw_seq[x: x + line_length] for x in range(0, len(raw_seq), line_length)):
            new_seq += j + "\n"

        self.seq = new_seq

        return new_seq

    def sub(self, site_start, site_end, strand, RNA):
        return sub_seq(self.seq, site_start, site_end, strand, RNA)

    def RNA(self):
        return self.seq.replace("T", "U")

    def reverse(self):
        return reverse_seq(self.seq)

    def complement(self):
        return complement_seq(self.seq)

    def reverse_complement(self):
        return complement_seq(reverse_seq(self.seq))

    def write_to_file(self, output_file, wrap_length=75):
        raw_seq = str(self.seq)
        with open(output_file, 'a') as f:
            f.write(">%s\n" % self.seqname)
            for x in range(0, len(raw_seq), wrap_length):
                # print(self.seqname, x)
                i = raw_seq[x: x + wrap_length]
                f.write(i + "\n")

    # def write_to_file(self, output_file, wrap_length=75):
    #     print(self.seqname, len(self.seq))
    #     self.wrap(line_length=wrap_length)
    #     print('ok wrap')
    #     # self.seq = re.sub("\n", "", self.seq)
    #     with open(output_file, 'a') as f:
    #         f.write(">%s\n%s" % (self.seqname, self.seq))


class FastaRecord(BioSeq):
    def __init__(self, seqname, seq=None, faidx=None):
        super(FastaRecord, self).__init__(seq)
        self.seqname = seqname
        self.faidx = faidx
        self.qualifiers = re.findall('\[([^\[\]]+)=([^\[\]]+)\]', seqname)

    def seqname_short(self):
        name_short = re.search('^(\S+)', self.seqname).group(1)
        return name_short

    @property
    def seq(self):
        if self.faidx is not None:
            return str(self.faidx)
        else:
            return self._seq

    @seq.setter
    def seq(self, value):
        self._seq = value

    def read_seq_to_mem(self):
        self._seq = str(self.faidx)

    def sub(self, site_start, site_end, strand, RNA):
        if self.faidx is not None:
            output = str(self.faidx[site_start - 1:site_end])
            if RNA is True:
                output = output.replace("T", "U")
            if strand == "-":
                output = reverse_complement(output, RNA)
            return output
        else:
            return sub_seq(self.seq, site_start, site_end, strand, RNA)

    def len(self):
        if hasattr(self, 'faidx'):
            return len(self.faidx)
        else:
            return len(self.seq)

    def get_GC_ratio(self):
        gc_now = (self.seq.count('C') + self.seq.count('G') +
                  self.seq.count('c') + self.seq.count('g'))/self.len()
        return gc_now


class FastqRecord(FastaRecord):
    def __init__(self, seqname, seq, quality):
        super(FastqRecord, self).__init__(seqname, seq)
        self.quality = quality


def read_fastq_big(file_name1, file_name2=None, gzip_flag=False):
    file_handle = []
    if gzip_flag is False:
        f1 = open(file_name1, 'r')
    else:
        f1 = gzip.open(file_name1, 'rt')
    file_handle.append(f1)
    if not file_name2 is None:
        if gzip_flag is False:
            f2 = open(file_name2, 'r')
        else:
            f2 = gzip.open(file_name2, 'rt')
        file_handle.append(f2)

    done = 0
    num = 0
    seqname = []
    seqs = []
    seqquality = []
    while not done:
        num = num + 1
        for f in file_handle:
            line_tmp = re.sub(r'\n', '', f.readline())

            if line_tmp == "":
                done = 1
                continue

            if num % 4 == 1:
                line_tmp = re.match(r'@(.*)', line_tmp).groups()[0]
                if re.match(r'^(.*)/1$', line_tmp):
                    line_tmp = re.match(r'^(.*)/1$', line_tmp).groups()[0]
                if re.match(r'^(.*)/2$', line_tmp):
                    line_tmp = re.match(r'^(.*)/2$', line_tmp).groups()[0]
                seqname.append(line_tmp)
            elif num % 4 == 2:
                seqs.append(line_tmp)
            elif num % 4 == 0:
                seqquality.append(line_tmp)

        if num % 4 == 0:
            if not file_name2 is None:
                f1_record = FastqRecord(seqname[0], seqs[0], seqquality[0])
                f2_record = FastqRecord(seqname[1], seqs[1], seqquality[1])
                yield (f1_record, f2_record)
            else:
                f1_record = FastqRecord(seqname[0], seqs[0], seqquality[0])
                yield (f1_record,)
            seqname = []
            seqs = []
            seqquality = []

    for f in file_handle:
        f.close()


def read_fasta(file_name, upper=True, filter_speci_char=False, gzip_flag=False, full_name=False):
    seqdict = {}
    seqname_list = []
    if gzip_flag is False:
        f = open(file_name, 'r')
    else:
        f = gzip.open(file_name, 'rt')
    all_text = f.read()
    all_text = re.sub('\r\n', '\n', all_text)
    # info = string.split(all_text, '>') python2
    info = all_text.split('\n>')
    while '' in info:
        info.remove('')
    for i in info:
        # seq = string.split(i, '\n', 1) python2
        seq = i.split('\n', 1)
        seq[1] = re.sub(r'\n', '', seq[1])
        seq[1] = re.sub(r' ', '', seq[1])
        seqname = seq[0]
        seqname = re.sub(r'^>', '', seqname)
        if filter_speci_char is True:
            seqname = re.search('^(\S+)', seqname).group(1)
            seqname = re.sub(r'[^A-Za-z0-9]', '_', seqname)
        name_short = re.search('^(\S+)', seqname).group(1)
        seqs = seq[1]
        if full_name:
            seqdict[seqname] = FastaRecord(seqname, seqs)
        else:
            seqdict[name_short] = FastaRecord(seqname, seqs)
        if upper is True:
            seqdict[name_short].seqs_upper()
        seqname_list.append(seqname)
    f.close()
    return seqdict, seqname_list


def read_fasta_big(file_name, upper=True, filter_speci_char=False, gzip_flag=False, log_file=None):
    module_log = logging_init("read_fasta_big", log_file)
    module_log.info('received a call to "lib.file_parser.read_fasta_big"')

    last_record_name = ""
    last_record_seq = ""
    top = 1

    if gzip_flag is False:
        f = open(file_name, 'r')
    else:
        f = gzip.open(file_name, 'rt')

    for each_line in f:
        each_line = re.sub(r'\n', '', each_line)
        match = re.match(r'^>(.*)', each_line)
        if match:
            seqname = match.group(1)
            # print(seqname)
            # short_name = re.search('^(\S+)', seqname).group(1)
            if top == 1:
                top = 2
                last_record_name = seqname
                if filter_speci_char is True:
                    last_record_name = re.search('^(\S+)', seqname).group(1)
                    last_record_name = re.sub(
                        r'[^A-Za-z0-9]', '_', last_record_name)
                last_record_seq = ""
                continue
            else:
                new_record = FastaRecord(last_record_name, last_record_seq)

                last_record_name = seqname
                if filter_speci_char is True:
                    last_record_name = re.search('^(\S+)', seqname).group(1)
                    last_record_name = re.sub(
                        r'[^A-Za-z0-9]', '_', last_record_name)
                last_record_seq = ""

                if upper is True:
                    new_record.seqs_upper()
                yield new_record

        else:
            each_line = re.sub(r'\n', '', each_line)
            last_record_seq = last_record_seq + each_line

    f.close()

    new_record = FastaRecord(last_record_name, last_record_seq)
    if upper is True:
        new_record.seqs_upper()
    yield new_record

    del module_log.handlers[:]


def faidx2xyx(pyfaidx_record, write_seq_mem):
    try:
        seq_name = pyfaidx_record.long_name
    except:
        seq_name = pyfaidx_record.name

    if write_seq_mem:
        return FastaRecord(seq_name, str(pyfaidx_record), pyfaidx_record)
    else:
        return FastaRecord(seq_name, None, pyfaidx_record)


def bioidx2xyx(Biopython_SeqRecord, write_seq_mem):
    try:
        seq_name = Biopython_SeqRecord.description
    except:
        seq_name = Biopython_SeqRecord.id

    if write_seq_mem:
        return FastaRecord(seq_name, str(Biopython_SeqRecord.seq), Biopython_SeqRecord)
    else:
        return FastaRecord(seq_name, None, Biopython_SeqRecord)


def wrap_fasta_file(file_name, wrap_file_name, wrap_length=75):
    with open(wrap_file_name, 'w') as w:
        for record_tmp in read_fasta_big(file_name, upper=False):
            w.write(">%s\n" % record_tmp.seqname)
            for i in (record_tmp.seq[x: x + wrap_length] for x in range(0, len(record_tmp.seq), wrap_length)):
                w.write("%s\n" % i)


def read_fasta_by_faidx(file_name, write_seq_mem=False, force_wrap_file=False):
    # if file_name.split(".")[-1] == 'gz':
    #     cmd_run("gzip %s" % file_name, slience=True)
    #     file_name = remove_file_name_suffix(file_name, 1)

    try:
        output_dict = OrderedDict()
        for i in Fasta(file_name):
            output_dict[i.name] = faidx2xyx(i, write_seq_mem)
        return output_dict
    except:
        try:
            cmd_run("samtools faidx %s" % file_name)
            output_dict = OrderedDict()
            for i in Fasta(file_name):
                output_dict[i.name] = faidx2xyx(i, write_seq_mem)
            return output_dict
        except:
            raise EnvironmentError("fasta file not suit for pyfaidx")

    #
    #
    # wrap_file_name = remove_file_name_suffix(file_name, 1) + ".pyfaidx.fasta"
    # if force_wrap_file:
    #     wrap_fasta_file(file_name, wrap_file_name, 75)
    #     try:
    #         return {i.name: faidx2xyx(i, write_seq_mem) for i in Fasta(wrap_file_name)}
    #     except:
    #         raise EnvironmentError("fasta file not suit for pyfaidx")
    # else:
    #     if os.path.exists(wrap_file_name):
    #         try:
    #             return {i.name: faidx2xyx(i, write_seq_mem) for i in Fasta(wrap_file_name)}
    #         except:
    #             wrap_fasta_file(file_name, wrap_file_name, 75)
    #             try:
    #                 return {i.name: faidx2xyx(i, write_seq_mem) for i in Fasta(wrap_file_name)}
    #             except:
    #                 raise EnvironmentError("fasta file not suit for pyfaidx")
    #     else:
    #         try:
    #             return {i.name: faidx2xyx(i, write_seq_mem) for i in Fasta(file_name)}
    #         except:
    #             wrap_fasta_file(file_name, wrap_file_name, 75)
    #             try:
    #                 return {i.name: faidx2xyx(i, write_seq_mem) for i in Fasta(wrap_file_name)}
    #             except:
    #                 raise EnvironmentError("fasta file not suit for pyfaidx")
    #


def read_fasta_by_bioidx(file_name, write_seq_mem=False):
    # if file_name.split(".")[-1] == 'gz':
    #     cmd_run("gzip %s" % file_name, slience=True)
    #     file_name = remove_file_name_suffix(file_name, 1)

    index_file = file_name + ".idx"

    bioidx_dict = SeqIO.index_db(index_file, file_name, 'fasta')
    output_dict = {}
    for i in bioidx_dict:
        try:
            seq_name = bioidx_dict[i].description
        except:
            seq_name = bioidx_dict[i].id

        output_dict[seq_name] = bioidx2xyx(bioidx_dict[i], write_seq_mem)

    return bioidx_dict


def read_fasta_to_sqlite(fasta_file_name, db_name, gzip_flag, log_file=None):
    module_log = logging_init("read_fasta_to_sqlite", log_file)
    module_log.info(
        'received a call to "lib.file_parser.read_fasta_to_sqlite"')

    # making a new sql database for store sequences'
    module_log.info('making a new sql database for store sequences')
    table_columns_dict = {
        "record": ["id", "seqname_short", "seqname", "seqs", "seqslen"]
    }
    sc.init_sql_db_many_table(db_name, table_columns_dict)
    module_log.info('made a new sql database for store sequences')

    # loading fasta file and store to sqlite
    module_log.info('loading fasta file and storing in sqlite')
    start_time = time.time()
    num = 0
    record_tmp_dict = []
    for record in read_fasta_big(fasta_file_name, upper=False, filter_speci_char=False, gzip_flag=gzip_flag):
        record_tmp_dict.append((num, record.seqname_short(
        ), record.seqname, record.seq, record.seqs_length()))
        num = num + 1

        if num % 10000 == 0:
            sc.sqlite_write(record_tmp_dict, db_name, "record",
                            table_columns_dict["record"])
            record_tmp_dict = []

        round_time = time.time()
        if round_time - start_time > 10:
            module_log.info("\tparsed: %d" % (num))
            start_time = round_time

    if len(record_tmp_dict) > 0:
        sc.sqlite_write(record_tmp_dict, db_name, "record",
                        table_columns_dict["record"])
        record_tmp_dict = []
        module_log.info("\tparsed: %d" % (num))

    module_log.info('loaded fasta file and stored in sqlite ')
    del module_log.handlers[:]


def extract_seq_from_sqlite(seq_name_list, db_name):
    records_info = sc.sqlite_select(
        db_name, 'record', key_name="seqname_short", value_tuple=seq_name_list)

    record_dict = {}
    for i in records_info:
        record_dict[i[1]] = FastaRecord(i[2], i[3])

    return record_dict


def split_fasta(file_or_dict_name, split_num, output_dir, file=1):
    if file == 1:
        seqdict, seqname_list = read_fasta(file_or_dict_name)
    else:
        seqdict, seqname_list = file_or_dict_name
    num = 0
    for i in (seqname_list[x: x + split_num] for x in range(0, len(seqname_list), split_num)):
        num = num + 1
        with open(output_dir + "/" + str(num) + ".seq", 'w') as f:
            for seq in i:
                seq_rec = seqdict[seq]
                seq_rec.wrap()
                f.write(">" + seq_rec.seqname + "\n" + seq_rec.seq)


def extract_seq_to_fasta(path_file_name, list1, fasta_dict, gzip_flag=False, silence=True):
    num = 0
    if gzip_flag is True:
        OUT = gzip.open(path_file_name, 'wt', newline='')
    else:
        OUT = open(path_file_name, 'w', newline='')
    for i in list1:
        if i in fasta_dict:
            fasta_dict[i].wrap()
            seq_now = fasta_dict[i].seqs
            printer = ">" + i + "\n" + seq_now + "\n"
            OUT.write(printer)
            num = num + 1
    OUT.close()
    if not silence:
        return num


def rename_seq_to_fasta(path_file_name, name_map_list, fasta_dict, gzip_flag=False, silence=True):
    num = 0
    if gzip_flag is True:
        OUT = gzip.open(path_file_name, 'wt', newline='')
    else:
        OUT = open(path_file_name, 'w', newline='')
    for raw_name, new_name in name_map_list:
        if raw_name in fasta_dict:
            fasta_dict[raw_name].wrap()
            seq_now = fasta_dict[raw_name].seq
            seq_now = seq_now.replace("*", "")
            # printer = ">" + new_name + "\n" + seq_now + "\n"
            printer = ">" + new_name + "\n" + seq_now
            OUT.write(printer)
            num = num + 1
    OUT.close()
    if not silence:
        return num


def sub_fasta(genome_file, seq_name, site_start, site_end, strand="+", RNA=False):
    target_record = None
    for i in read_fasta_big(genome_file):
        if seq_name == i.seqname_short():
            target_record = i
            break

    if target_record:
        sub_seq_out = sub_seq(
            target_record.seq, site_start, site_end, strand, RNA)

        start = min(site_start, site_end)
        end = max(site_start, site_end)

        sub_seq_name = "%s:%d-%d:%s" % (seq_name, start, end, strand)
        record = FastaRecord(sub_seq_name, sub_seq_out)
        return record
    else:
        raise NameError('No target seq')


def sub_fasta_many(genome_file, bed_dict):
    """
    :param genome_file:
    :param bed_dict: a dict whoes keys are subseq names and values should be tuple as (seq_name, site_start, site_end, strand="+", RNA=False)
    :return:
    """
    bed_dict_seqname = OrderedDict()
    for sub_seq_name in bed_dict:
        seq_name, site_start, site_end, strand, RNA = bed_dict[sub_seq_name]
        if not seq_name in bed_dict_seqname:
            bed_dict_seqname[seq_name] = []
        bed_dict_seqname[seq_name].append(sub_seq_name)

    output_dict = OrderedDict()
    for record in read_fasta_big(genome_file):
        if not record.seqname_short() in bed_dict_seqname:
            continue
        for sub_seq_name in bed_dict_seqname[record.seqname_short()]:
            seq_name, site_start, site_end, strand, RNA = bed_dict[sub_seq_name]
            start = min(int(site_start), int(site_end))
            end = max(int(site_start), int(site_end))
            sub_seq = record.seq[start - 1:end]

            if strand == "-":
                sub_seq = reverse_complement(sub_seq)
            if RNA is True:
                sub_seq = sub_seq.replace("T", "U")

            output_dict[sub_seq_name] = FastaRecord(sub_seq_name, sub_seq)

    return output_dict


def aa_aln_to_cds_aln(aa_aln, cds_fasta, cds_aln):
    cmd_string = "treebest backtrans "+aa_aln+" "+cds_fasta+" > "+cds_aln
    cmd_run(cmd_string, silence=True)


# normal seq id parse

def fancy_name_parse(input_string):
    contig_name, c_start, c_end = re.search(
        r'^(\S+):(\d+)-(\d+)$', input_string).groups()
    return contig_name, int(c_start), int(c_end)


def fancy_name_get(contig_name, c_start, c_end):
    return "%s:%s-%s" % (contig_name, str(c_start), str(c_end))


# database seq parser tools

def record_qulifiers_dict(qualifiers):
    """
    this function work for NCBI fasta id
    :param qualifiers:
    :return:
    """
    output = {}
    for i in qualifiers:
        i = re.sub(r'\[', '', i)
        i = re.sub(r'\]', '', i)
        try:
            key_tmp, value_tmp = i.split("=")
            output[key_tmp] = value_tmp
        except:
            pass
    return output


def NCBI_id_parse(nt_file, gzip_flag=False):
    """
    # gene id in ncbi gff hard to link to pt cds fasta id, so we need change to locus_tag or Dbxref and so on

    parse NCBI seq fasta file, to locus_tag or Dbxref as seq ID
    :param nt_file:
    :param gzip_flag:
    :return:
    """

    seqdict, seqname_list = read_fasta(nt_file, gzip_flag=gzip_flag)
    qualifiers_dict = {}
    seqdict_new = {}
    for i in seqdict:
        record = seqdict[i]
        qualifiers = re.findall(r'\[\S+\]', record.seqname)
        qualifiers = record_qulifiers_dict(qualifiers)
        if 'pseudo' in qualifiers and qualifiers['pseudo'] == 'true':
            continue

        prot_id = re.findall(r'^lcl\|(\S+)_prot_(\S+)$', i)
        cds_id = re.findall(r'^lcl\|(\S+)_cds_(\S+)$', i)

        if len(prot_id) > 0:
            contig_id = prot_id[0][0]
        elif len(cds_id) > 0:
            contig_id = cds_id[0][0]
        else:
            raise ValueError("new type fasta? %s" % nt_file)

        qualifiers['raw_id'] = i
        qualifiers['contig'] = contig_id

        if (contig_id, qualifiers['location']) in qualifiers_dict:
            if not seqdict_new[(contig_id, qualifiers['location'])].seq == record.seq:
                # I find in GCF_000001215, gene CG33301 have two protein from same CDS, but one AA to be *
                # if len(re.sub('\*','',seqdict_new[(contig_id, qualifiers['location'])].seq)) < len(re.sub('\*','',record.seq)):
                #     seqdict_new[(contig_id, qualifiers['location'])] = record
                #
                if contig_id == 'NT_033779.5' and qualifiers[
                        'location'] == 'join(10049616..10050642,10050702..10050898)':
                    if len(re.sub('\*', '', seqdict_new[(contig_id, qualifiers['location'])].seq)) < len(
                            re.sub('\*', '', record.seq)):
                        seqdict_new[(
                            contig_id, qualifiers['location'])] = record
                    else:
                        continue
                else:
                    raise ValueError("same seq id but diff seq: %s %s" % (
                        record.seqname, nt_file))
        else:
            qualifiers_dict[(contig_id, qualifiers['location'])] = qualifiers
            record.seqname = i
            seqdict_new[(contig_id, qualifiers['location'])] = record

        # """use protein_id as gene id"""
        # #
        # # if 'protein_id' in qualifiers:
        # #     if qualifiers['protein_id'] in qualifiers_dict:
        # #         del qualifiers_dict[qualifiers['protein_id']]
        # #         del seqdict_new[qualifiers['protein_id']]
        # #     else:
        # #         qualifiers_dict[qualifiers['protein_id']] = qualifiers
        # #         record.seqname = qualifiers['protein_id']
        # #         seqdict_new[qualifiers['protein_id']] = record
        #
        # if qualifiers[renew_id_flag] in qualifiers_dict:
        #     if not seqdict_new[qualifiers[renew_id_flag]].seq == record.seq:
        #         raise ValueError("same seq id but diff seq: %s %s" % (record.seqname, nt_file))
        # else:
        #     qualifiers_dict[qualifiers[renew_id_flag]] = qualifiers
        #     record.seqname = qualifiers[renew_id_flag]
        #     seqdict_new[qualifiers[renew_id_flag]] = record
        #
        # """go back to locus_tag"""
        #
        # # if 'locus_tag' in qualifiers:
        # #     qualifiers_dict[qualifiers['locus_tag']] = qualifiers
        # #     record.seqname = qualifiers['locus_tag']
        # #     seqdict_new[qualifiers['locus_tag']] = record

    return qualifiers_dict, seqdict_new


def JGI_id_parse(file, gzip_flag=False):
    seqdict, seqname_list = read_fasta(file, gzip_flag=gzip_flag)
    qualifiers_dict = {}
    seqdict_new = {}
    for i in seqdict:
        record = seqdict[i]
        qualifiers = record.seqname.split("|", 3)
        spec_tag = qualifiers[1]
        seq_id = qualifiers[2]
        seq_name = qualifiers[3]
        seqdict_new[seq_name] = record
        qualifiers_dict[seq_name] = {}
        qualifiers_dict[seq_name]["seq_id"] = seq_id
        qualifiers_dict[seq_name]["spec_tag"] = spec_tag
        qualifiers_dict[seq_name]["seq_name"] = seq_name
    return qualifiers_dict, seqdict_new


def BIG_id_parse(nt_file, gzip_flag=False):
    """
    # gene id in ncbi gff hard to link to pt cds fasta id, so we need change to locus_tag or Dbxref and so on

    parse NCBI seq fasta file, to locus_tag or Dbxref as seq ID
    :param nt_file:
    :param gzip_flag:
    :return:
    """

    seqdict, seqname_list = read_fasta(nt_file, gzip_flag=gzip_flag)
    qualifiers_dict = {}
    seqdict_new = {}
    for i in seqdict:
        record = seqdict[i]

        # parse ID info
        qualifiers = {}
        record_info = record.seqname.split('\t')
        qualifiers['raw_id'] = record_info[0]
        for j in record_info[1:]:
            name, value = re.findall(r'^(\S+)=(.*)$', j)[0]
            qualifiers[name] = value

        if 'Protein' in qualifiers:
            uniq_name = qualifiers['Protein']
        else:
            uniq_name = qualifiers['raw_id']

        qualifiers_dict[uniq_name] = qualifiers
        record.seqname = uniq_name
        seqdict_new[uniq_name] = record

    return qualifiers_dict, seqdict_new


def PreExtractNr(nr_fasta_file, sql3_db_file):
    import time
    import sqlite3
    from lib.xuyuxing.base.common_command import log_print
    import lib.common.sqlite_command as sc

    start_time = time.time()
    table_columns_dict = {
        "name_record": ("accession_id", "annotation_string", "record_id"),
        "seq_record": ("record_id", "seq")
    }
    sc.init_sql_db_many_table(sql3_db_file, table_columns_dict)

    record_id = 0
    name_record = []
    seq_record = []
    for record in read_fasta_big(nr_fasta_file, upper=True, filter_speci_char=False, gzip_flag=False):
        record_id = record_id + 1
        huge_name = record.seqname
        seq = record.seq
        for name in re.split('\x01', huge_name):
            accession_id = name.split(" ", 1)[0]
            # print("acc:%s\tfunc:%s\tspeci:%s" % (accession_id, function_string, speci_name))
            name_record.append((accession_id, name, record_id))
        seq_record.append((record_id, seq))

        if record_id % 10000 == 0:
            sc.sqlite_write(name_record, sql3_db_file,
                            "name_record", table_columns_dict["name_record"])
            name_record = []
            sc.sqlite_write(seq_record, sql3_db_file,
                            "seq_record", table_columns_dict["seq_record"])
            seq_record = []

        round_time = time.time()
        if round_time - start_time > 10:
            log_print("%d finished" % (record_id))
            start_time = round_time

    sc.sqlite_write(name_record, sql3_db_file, "name_record",
                    table_columns_dict["name_record"])
    sc.sqlite_write(seq_record, sql3_db_file, "seq_record",
                    table_columns_dict["seq_record"])
    log_print("%d finished" % (record_id))

    conn = sqlite3.connect(sql3_db_file)
    conn.execute("CREATE UNIQUE INDEX name_index on %s (\"%s\")" %
                 ("name_record", "accession_id"))
    conn.execute("CREATE UNIQUE INDEX seq_index on %s (\"%s\")" %
                 ("seq_record", "record_id"))
    conn.close()


def ExtractNr(nr_fasta_file, key_IDs, output_file, short_name=True):
    from lib.common.sqlite_command import sqlite_select_by_a_key
    name_IDs = sqlite_select_by_a_key(
        nr_fasta_file, "name_record", "accession_id", key_IDs)
    seq_IDs = tuple(set([i[2] for i in name_IDs]))
    seq_IDs = sqlite_select_by_a_key(
        nr_fasta_file, "seq_record", "record_id", seq_IDs)
    record_dict = []
    for accession_id, annotation_string, record_id in name_IDs:
        seqs = [i[1] for i in seq_IDs if i[0] == record_id][0]
        if short_name is True:
            seqname = accession_id
        else:
            seqname = annotation_string
        record_dict.append(FastaRecord(seqname, seqs))

    if output_file is not None:
        F1 = open(output_file, 'w')

    record_dict = list(set(record_dict))

    for record in record_dict:
        record.wrap()
        if output_file is not None:
            F1.write(">%s\n%s\n" % (record.seqname, record.seq))
        else:
            print(">%s\n%s" % (record.seqname, record.seq))
    if output_file is not None:
        F1.close()

    return len(record_dict), len(key_IDs)


def seq_cleaner(fasta_seq_record, seq_type='prot', degenerate_sites_allowed=False, gap_allowed=False,
                translation_stop_allowed=False, replace_flag=False, full_name_flag=False):
    """
    usage: SeqParser FastaFormatClean [-h] [-t {prot,nucl}] [-d] [-g] [-s]
                                  [-w WRAP] [-r] [-n] [-l LOG_FILE]
                                  raw_fasta_file output_fasta_file

    positional arguments:
      raw_fasta_file        raw fasta file
      output_fasta_file     output fasta file

    optional arguments:
      -h, --help            show this help message and exit
      -t {prot,nucl}, --seq_type {prot,nucl}
                            what type sequences you have? (default:prot)
      -d, --degenerate_sites_allowed
                            allow degenerate sites
      -g, --gap_allowed     allow gap sites
      -s, --translation_stop_allowed
                            allow translation stop sites
      -w WRAP, --wrap WRAP  all lines of text be shorter than warp characters in
                            length
      -r, --replace_flag    if find a unknown or not allowed site, should we
                            replace to any site? (default: no and skip sequence)
      -n, --full_name_flag  use full sequence name in output
      -l LOG_FILE, --log_file LOG_FILE
                            path for log file (default:None)

    """

    # argument parser
    if seq_type == 'nucl':
        seq_type = 0b0001
    elif seq_type == 'prot':
        seq_type = 0b0000

    dsa_flag = degenerate_sites_allowed
    if dsa_flag:
        dsa_flag = 0b0010

    ga_flag = gap_allowed
    if ga_flag:
        ga_flag = 0b0100

    tsa_flag = translation_stop_allowed
    if tsa_flag:
        tsa_flag = 0b1000

    bit_field = seq_type + dsa_flag + ga_flag + tsa_flag

    if bit_field == 0b0111:
        all_allowed = FastaRecord.nucl + FastaRecord.any_nucl + \
            FastaRecord.nucl_degenerate + FastaRecord.gap_site
        safe_remove = []
        safe_replace = FastaRecord.any_nucl
    elif bit_field == 0b0011:
        all_allowed = FastaRecord.nucl + FastaRecord.any_nucl + FastaRecord.nucl_degenerate
        safe_remove = FastaRecord.gap_site
        safe_replace = FastaRecord.any_nucl
    elif bit_field == 0b0101:
        all_allowed = FastaRecord.nucl + FastaRecord.any_nucl + FastaRecord.gap_site
        safe_remove = []
        safe_replace = FastaRecord.any_nucl
    elif bit_field == 0b0001:
        all_allowed = FastaRecord.nucl + FastaRecord.any_nucl
        safe_remove = FastaRecord.gap_site
        safe_replace = FastaRecord.any_nucl
    elif bit_field == 0b0000:
        all_allowed = FastaRecord.prot + FastaRecord.any_prot
        safe_remove = FastaRecord.gap_site + FastaRecord.stop_site
        safe_replace = FastaRecord.any_prot
    elif bit_field == 0b0100:
        all_allowed = FastaRecord.prot + FastaRecord.any_prot + FastaRecord.gap_site
        safe_remove = FastaRecord.stop_site
        safe_replace = FastaRecord.any_prot
    elif bit_field == 0b1100:
        all_allowed = FastaRecord.prot + FastaRecord.any_prot + \
            FastaRecord.gap_site + FastaRecord.stop_site
        safe_remove = []
        safe_replace = FastaRecord.any_prot
    elif bit_field == 0b1000:
        all_allowed = FastaRecord.prot + FastaRecord.any_prot + FastaRecord.stop_site
        safe_remove = FastaRecord.gap_site
        safe_replace = FastaRecord.any_prot
    else:
        raise ValueError("Bad input flag")

    fasta_seq_record.no_wrap()
    new_seq = fasta_seq_record.seq
    char_site = set(new_seq)
    bad_flag = False
    for i in char_site:
        if i in all_allowed:
            continue
        elif i in safe_remove:
            if i == "*":
                new_seq = re.sub("\*", "", new_seq)
            else:
                new_seq = re.sub(i, "", new_seq)
        else:
            if replace_flag:
                if i == "*":
                    new_seq = re.sub("\*", "", new_seq)
                else:
                    new_seq = re.sub(i, safe_replace[0], new_seq)
            else:
                bad_flag = True

    if bad_flag is True:
        good_flag = False
    else:
        good_flag = True
        fasta_seq_record.seq = new_seq

        if full_name_flag:
            fasta_seq_record.seqname = fasta_seq_record.seqname
        else:
            fasta_seq_record.seqname = fasta_seq_record.seqname_short()

    return good_flag, fasta_seq_record


def write_fasta(fasta_record_list, output_file, wrap_length=75, upper=False):
    with open(output_file, 'w') as f:
        for record in fasta_record_list:
            # print(record.seqname)
            # print(record.seq)
            record.no_wrap()
            raw_seq = str(record.seq)

            if upper:
                raw_seq = raw_seq.upper()

            f.write(">%s\n" % record.seqname)
            for x in range(0, len(raw_seq), wrap_length):
                # print(self.seqname, x)
                i = raw_seq[x: x + wrap_length]
                f.write(i + "\n")


if __name__ == '__main__':
    fasta_file = "/lustre/home/xuyuxing/Work/Other/saif/jupyter/A188.rename.fa"
    seq_dict = read_fasta_by_faidx(fasta_file)
    seq_tmp = seq_dict['CHR1']
    sub_seq_tmp = seq_tmp.sub(1, 100, '+', RNA=False)
    sub_seq_tmp = BioSeq(sub_seq_tmp)
    sub_seq_tmp.reverse_complement()

    # from Bio import SeqIO
    #
    # from lib.base.base_function import cmd_run
    #
    # # build
    # a = SeqIO.index_db('/lustre/home/xuyuxing/Work/Other/saif/jupyter/A188.rename.fa.gz.idx',
    #                    '/lustre/home/xuyuxing/Work/Other/saif/jupyter/A188.rename.fa.gz', 'fasta')
    # a['CHR1'][1000:100000]
    # seq_dict={}
    # for i in a:
    #     seq_dict[i]=a[i]
    #
    # seq_dict = read_fasta_by_bioidx(fasta_file)



# Do not change!
TAXURL = 'http://rest.kegg.jp/list/organism'
KOURL = 'http://rest.kegg.jp/link/ko/'
ECURL = 'http://rest.kegg.jp/link/enzyme/'

PWURL = 'http://rest.kegg.jp/link/pathway/'

ECURL = 'http://rest.kegg.jp/link/enzyme/'
ECURL = 'http://rest.kegg.jp/link/enzyme/'
ECURL = 'http://rest.kegg.jp/link/enzyme/'





if __name__ == '__main__':

    import argparse

    # command argument parse

    # argument parse
    parser = argparse.ArgumentParser(
        prog='GenomeAnnotation', description='tools help to get genome annotation more easy\n'
    )

    subparsers = parser.add_subparsers(
        title='subcommands', dest="subcommand_name")

    # argparse for ExtractGFF
    parser_a = subparsers.add_parser('ExtractGFF',
                                     help='extract genome feature record from gff file')

    parser_a.add_argument("genome_feature_id_list",
                          help="genome_feature_id_list", type=str)
    parser_a.add_argument("raw_gff_file", help="raw gff file", type=str)
    parser_a.add_argument(
        "output_gff_file", help="path of output gff", type=str)

    # argparse for GFFRename
    parser_a = subparsers.add_parser('GFFRename',
                                     help='Rename a gff file')

    parser_a.add_argument(
        "input_gff_file", help="gff file for input", type=str)
    parser_a.add_argument(
        "output_gff_file", help="gff file for output", type=str)
    parser_a.add_argument(
        "rename_prefix", help="gene name rename prefix", type=str)
    parser_a.add_argument("-c", "--rename_chr_id",
                          help="rename_chr_id", action='store_true')

    # argparse for SortGFF
    parser_a = subparsers.add_parser('SortGFF',
                                     help='sort a gff file by feature location')

    parser_a.add_argument(
        "input_gff_file", help="gff file for input", type=str)
    parser_a.add_argument(
        "output_gff_file", help="gff file for output", type=str)
    parser_a.add_argument("-t", "--sort_for_tabix",
                          help="sort gff file for tabix", action='store_true')

    # argparse for FeatureLen
    parser_a = subparsers.add_parser('FeatureLen',
                                     help='get model mRNA length from a gff file')

    parser_a.add_argument(
        "input_gff_file", help="gff file for input", type=str)

    # argparse for GffModel
    parser_a = subparsers.add_parser('GffModel',
                                     help='get gene model from NCBI gff',
                                     description='get gene model from NCBI gff by the longest transcript')

    parser_a.add_argument('ncbi_gff', type=str,
                          help='Path of NCBI genomic gff')
    parser_a.add_argument("-o", "--output_file", help="Output file name (default as *.gene.model)", default=None,
                          type=str)

    # argparse for NtPtMap
    parser_b = subparsers.add_parser('NtPtMap',
                                     help='find the gene which have pt and nt both',
                                     description='for mission which need use pt and nt both')

    parser_b.add_argument('Nt_file', type=str,
                          help='Path of Nt file with NCBI fasta format')
    parser_b.add_argument('Pt_file', type=str,
                          help='Path of Pt file with NCBI fasta format')
    parser_b.add_argument('gene_model', type=str,
                          help='Path of gene mode file from GffModel')
    parser_b.add_argument('tag', type=str, help='tag for output')

    args = parser.parse_args()
    args_dict = vars(args)

    # ---------------------------------------------------------
    # command detail

    if args_dict["subcommand_name"] == "ExtractGFF":
        from src.xuyuxing.tools.gene_feature_tools import extract_gff_main
        extract_gff_main(args)

    elif args_dict["subcommand_name"] == "GFFRename":
        from src.xuyuxing.tools.gene_feature_tools import gff_rename_main
        gff_rename_main(args)

    elif args_dict["subcommand_name"] == "FeatureLen":
        from lib.common.genome.genome_feature2 import read_gff_file, Gene

        gf_dict = read_gff_file(args.input_gff_file)

        for feature_tmp in gf_dict:
            for gf_id in gf_dict[feature_tmp]:
                gf = gf_dict[feature_tmp][gf_id]
                gene = Gene(from_gf=gf)
                gene.build_gene_seq()
                if 'exon' in gene.model_mRNA.sgf_len_dir:
                    print("%s\t%s" %
                          (gf.id, gene.model_mRNA.sgf_len_dir['exon']))
                else:
                    print("%s\t%s" %
                          (gf.id, gene.model_mRNA.sgf_len_dir['CDS']))

    elif args_dict["subcommand_name"] == "SortGFF":
        from src.xuyuxing.tools.gene_feature_tools import sort_gff_file
        sort_gff_file(args)

    elif args_dict["subcommand_name"] == "GffModel":
        from lib.common.genome.seq_base import read_fasta
        from BCBio import GFF
        import re

        in_file = args.ncbi_gff
        output_file = args.output_file

        cmd_string = "sed -i '/;part=/d' " + in_file
        cmd_run(cmd_string)

        in_handle = open(in_file)
        gene_dict = {}
        num = 0
        gene_model = {}
        for rec in GFF.parse(in_handle):
            for gene in rec.features:
                gene_dict[gene.id] = gene, rec.id
                if 'gene_biotype' in gene.qualifiers and gene.qualifiers['gene_biotype'][0] == 'protein_coding':
                    protein_dict = {}
                    for rna in gene.sub_features:
                        for cds in rna.sub_features:
                            if cds.type == "CDS":
                                protein_id = cds.qualifiers["protein_id"][0]
                                cds_length = abs(
                                    cds.location.start - cds.location.end)
                                if protein_id not in protein_dict:
                                    protein_dict[protein_id] = cds_length
                                else:
                                    protein_dict[protein_id] = protein_dict[protein_id] + cds_length
                    for cds in gene.sub_features:
                        if cds.type == "CDS":
                            protein_id = cds.qualifiers["protein_id"][0]
                            cds_length = abs(
                                cds.location.start - cds.location.end)
                            if protein_id not in protein_dict:
                                protein_dict[protein_id] = cds_length
                            else:
                                protein_dict[protein_id] = protein_dict[protein_id] + cds_length
                    model_protein = sorted(
                        protein_dict, key=lambda x: protein_dict[x], reverse=True)[0]
                    gene_model[gene.id] = model_protein
        in_handle.close()

        if output_file is not None:
            output_file = output_file
        else:
            output_file = in_file + ".gene.model"

        with open(output_file, 'w') as f:
            for i in gene_model:
                f.write(i + "\t" + gene_model[i] + "\n")

    elif args_dict["subcommand_name"] == "NtPtMap":
        from lib.common.genome.seq_base import read_fasta
        from api.xuyuxing.file_parser.fileIO import tsv_file_parse
        from BCBio import GFF
        import re

        def record_qulifiers_dict(qualifiers):
            output = {}
            for i in qualifiers:
                i = re.sub(r'\[', '', i)
                i = re.sub(r'\]', '', i)
                try:
                    key_tmp, value_tmp = i.split("=")
                    output[key_tmp] = value_tmp
                except:
                    pass
            return output

        def NCBI_nt_id_parse(nt_file):
            seqdict, seqname_list = read_fasta(nt_file)
            qualifiers_dict = {}
            seqdict_new = {}
            for i in seqdict:
                record = seqdict[i]
                qualifiers = re.findall(r'\[\S+\]', record.seqname)
                qualifiers = record_qulifiers_dict(qualifiers)
                if 'pseudo' in qualifiers and qualifiers['pseudo'] == 'true':
                    continue
                qualifiers['raw_id'] = i
                if qualifiers['protein_id'] in qualifiers_dict:
                    del qualifiers_dict[qualifiers['protein_id']]
                    del seqdict_new[qualifiers['protein_id']]
                else:
                    qualifiers_dict[qualifiers['protein_id']] = qualifiers
                    record.seqname = qualifiers['protein_id']
                    seqdict_new[qualifiers['protein_id']] = record
            return qualifiers_dict, seqdict_new

        nt_file = args.Nt_file
        pt_file = args.Pt_file
        gene_model_file = args.gene_model
        tag = args.tag

        pt_seqdict, pt_seqname_list = read_fasta(pt_file)
        qualifiers_dict, nt_seqdict = NCBI_nt_id_parse(nt_file)

        Maped_id = set(pt_seqdict.keys()) & set(qualifiers_dict.keys())
        gene_mode = tsv_file_parse(gene_model_file)
        gene_mode = [gene_mode[id][1] for id in gene_mode.keys()]

        output_id = set(Maped_id) & set(gene_mode)

        with open(tag + ".fasta", 'w') as f:
            for i in output_id:
                record = pt_seqdict[i]
                f.write(">%s|%s\n%s\n" % (tag, i, record.seqs))

        with open(tag + ".nt.fasta", 'w') as f:
            for i in output_id:
                record = nt_seqdict[i]
                f.write(">%s|%s\n%s\n" % (tag, i, record.seqs))

        print("%s\t%d\t%d" % (tag, len(gene_mode), len(output_id)))



from config import gene_painter_path, clustalw_path
import uuid
from lib.common.os import mkdir, rmdir, cmd_run
import os
from lib.common.genome.genome_feature2 import write_gff_file, sub_gf_traveler, mRNA_rename, sub_gf_traveler
import re
from itertools import combinations


def read_genepainter_intron_phase_file(intron_file):
    """
    I can just read genepainter-intron-phase.txt, gene names should not be too long and have '-'
    """

    intron_dict = {}

    with open(intron_file, 'r') as f:
        for each_line in f:
            each_line = each_line.strip()
            info = each_line.split()
            g_name = re.sub('>', '', info[0])
            intron_aln = []
            for i in range(1, len(info[1]), 2):
                intron_aln.append(info[1][i])
            intron_dict[g_name] = intron_aln

    return intron_dict


def intron_map(mRNA_list, intron_dict):
    intron_mapped_dict = {}

    for mRNA_gf in mRNA_list:
        intron_mapped_dict[mRNA_gf] = []
        mRNA_gf.get_introns()
        intron_list = [i for i in sub_gf_traveler(
            mRNA_gf) if i.type == 'intron']
        sorted_intron_list = sorted(
            intron_list, key=lambda x: x.start, reverse=(mRNA_gf.strand == '-'))

        num = 0
        for i in intron_dict[mRNA_gf.id]:
            if i == '-':
                intron_mapped_dict[mRNA_gf].append(None)
            else:
                intron = sorted_intron_list[num]
                intron.phase = i
                intron_mapped_dict[mRNA_gf].append(intron)
                num += 1

    return intron_mapped_dict


def intron_aligner(mRNA_list, top_work_dir="/tmp"):
    """
    mRNA in mRNA_list should fit the object from genome_feature2, mRNA should have aa_seq
    """

    top_work_dir = os.path.abspath(top_work_dir)

    work_dir = top_work_dir + "/" + uuid.uuid1().hex
    mkdir(work_dir)

    # gene rename
    rename_dict = {}

    num = 0
    mRNA_rename_list = []
    mRNA_only_cds_list = []
    for mRNA_gf in mRNA_list:
        new_id = 'g%d' % num
        rename_dict[new_id] = mRNA_gf.id
        cds_list = [i for i in sub_gf_traveler(mRNA_gf) if i.type == 'CDS']
        mRNA_gf.sub_features = cds_list

        mRNA_only_cds_list.append(mRNA_gf)
        mRNA_gf = mRNA_rename(mRNA_gf, new_id)
        mRNA_rename_list.append(mRNA_gf)
        num += 1

    # write aa fasta
    aa_fasta_file = work_dir + "/aa.fasta"
    with open(aa_fasta_file, 'w') as f:
        for mRNA_gf in mRNA_rename_list:
            f.write(">%s\n%s\n" % (mRNA_gf.id, mRNA_gf.aa_seq))

    # running clustalw2 to get aln
    aa_aln_file = work_dir + "/aa.aln"
    cmd_string = "%s -INFILE=%s -ALIGN -OUTPUT=FASTA -OUTFILE=%s -type=protein" % (
        clustalw_path, aa_fasta_file, aa_aln_file)
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    # prepare gff dir
    gff_dir = work_dir + "/gff"
    mkdir(gff_dir)

    for mRNA_gf in mRNA_rename_list:
        write_gff_file([mRNA_gf], gff_dir+"/"+mRNA_gf.id +
                       ".gff3", source=None, sort=False)

    # run  gene_painter
    cmd_string = "%s -i aa.aln -p gff --intron-phase" % gene_painter_path
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    intron_phase_file = work_dir + "/genepainter-intron-phase.txt"

    # parse intron phase
    intron_dict = read_genepainter_intron_phase_file(intron_phase_file)

    rename_intron_dict = {}
    for i in intron_dict:
        rename_intron_dict[rename_dict[i]] = intron_dict[i]

    mapped_intron_dict = intron_map(mRNA_only_cds_list, rename_intron_dict)

    rmdir(work_dir)

    return mapped_intron_dict


def intron_length_diverse_index(intron_dict_from_intron_aligner):
    """
    key: gene_gf
    value: aligned intron gf
    """
    output_dict = {}
    for a, b in combinations(list(intron_dict_from_intron_aligner.keys()), 2):
        a_intron_list = intron_dict_from_intron_aligner[a]
        b_intron_list = intron_dict_from_intron_aligner[b]

        if len(a_intron_list) != len(b_intron_list):
            raise ValueError('intron list should aligned')
        elif len(a_intron_list) == 0:
            return 1.0
        else:
            intron_diverse_value_list = []
            for i in range(len(a_intron_list)):
                ai = a_intron_list[i]
                bi = b_intron_list[i]

                if ai is None and bi is not None:
                    intron_diverse_value = 0.0
                elif not ai is None and bi is None:
                    intron_diverse_value = 0.0
                elif ai is None and bi is None:
                    continue
                else:
                    intron_diverse_value = min(
                        ai.len(), bi.len())/max(ai.len(), bi.len())

                intron_diverse_value_list.append(intron_diverse_value)

            weigthed_intron_diverse_value = sum(
                intron_diverse_value_list)/len(intron_diverse_value_list)

            output_dict[(a, b)] = weigthed_intron_diverse_value
    
    return output_dict


if __name__ == "__main__":
    from lib.common.genome.genome_feature2 import read_gff_file, mRNA
    from lib.common.genome.seq_base import read_fasta_by_faidx

    Aof_gff_file = '/lustre/home/xuyuxing/Database/Plant_genome/clean_data/Asparagus_officinalis/T4686N0.genome.gff3'
    Aof_gff_dict = read_gff_file(Aof_gff_file)
    Aof_gene_dict = Aof_gff_dict['gene']

    Aof_aa_fasta_file = '/lustre/home/xuyuxing/Database/Plant_genome/clean_data/Asparagus_officinalis/T4686N0.gene_model.protein.fasta'
    Aof_aa_dict = read_fasta_by_faidx(Aof_aa_fasta_file)

    mRNA1 = mRNA(from_gf=Aof_gene_dict['T4686N0C00001G02624'].sub_features[0])
    mRNA2 = mRNA(from_gf=Aof_gene_dict['T4686N0C00003G01631'].sub_features[0])

    mRNA1.aa_seq = Aof_aa_dict['T4686N0C00001G02624'].seq
    mRNA2.aa_seq = Aof_aa_dict['T4686N0C00003G01631'].seq

    mRNA_list = [mRNA1, mRNA2]
    top_work_dir = '/lustre/home/xuyuxing/Work/Gel/synteny/20210311/orthofinder/pt_file/OrthoFinder/Results_Jun23/MultipleSequenceAlignments/test'
    intron_dict = intron_aligner(mRNA_list, top_work_dir)

    #
    from lib.common.evolution.orthotools2 import OrthoGroups, read_species_info
    from lib.common.genome.genome_feature2 import read_gff_file, mRNA
    from lib.common.genome.seq_base import read_fasta_by_faidx

    ref_xlsx = '/lustre/home/xuyuxing/Work/Gel/Gene_Loss2/S44_version3/Gel_ref.xlsx'
    OG_tsv_file = '/lustre/home/xuyuxing/Work/Gel/intron/orthofinder/pt_file/OrthoFinder/Results_Jun27/Orthogroups/Orthogroups.tsv'
    work_dir = '/lustre/home/xuyuxing/Work/Gel/intron/'

    sp_info_dict = read_species_info(ref_xlsx)

    all_data_dict = {}
    for sp_id in sp_info_dict:
        sp = sp_info_dict[sp_id]

        gff_dict = read_gff_file(sp.gff_file)
        gene_dict = gff_dict['gene']

        aa_dict = read_fasta_by_faidx(sp.pt_file)
        cds_dict = read_fasta_by_faidx(sp.cds_file)

        mRNA_dict = {}
        for g_id in gene_dict:
            gf = gene_dict[g_id]
            if g_id in aa_dict:
                mRNA_gf = mRNA(from_gf=gf.sub_features[0], aa_seq=aa_dict[g_id].seq, cds_seq=cds_dict[g_id].seq)
                mRNA_dict[mRNA_gf.id] = mRNA_gf
            
        all_data_dict[sp_id] = mRNA_dict

    OGs = OrthoGroups(OG_tsv_file=OG_tsv_file)
    aln_work_dir = work_dir + "/intron_aln"
    mkdir(aln_work_dir, True)

    og_id = 'OG0009838'
    work_dir_now = aln_work_dir + "/" + og_id
    mkdir(work_dir_now, True)
    og = OGs.OG_dict[og_id]

    mRNA_list = []
    for gene in og.gene_list:
        mRNA_list.append(all_data_dict[gene.species][gene.id])

    intron_aligner(mRNA_list, top_work_dir=work_dir_now)    
       
        





# coding: utf-8

# ipython notebook requires this
# %matplotlib inline

# python console requires this
import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
import api.xuyuxing.plot.VennDiagram as venn

labels = venn.get_labels([range(10), range(5, 15)], fill=['number', 'logic'])
fig, ax = venn.venn2(labels, names=['list 1', 'list 2'])
fig.savefig('venn2.png', bbox_inches='tight')
plt.close()

labels = venn.get_labels([range(10), range(5, 15), range(3, 8)], fill=['number', 'logic'])
fig, ax = venn.venn3(labels, names=['list 1', 'list 2', 'list 3'])
fig.savefig('venn3.png', bbox_inches='tight')
plt.close()

labels = venn.get_labels([range(10), range(5, 15), range(3, 8), range(8, 17)], fill=['number', 'logic'])
fig, ax = venn.venn4(labels, names=['list 1', 'list 2', 'list 3', 'list 4'])
fig.savefig('venn4.png', bbox_inches='tight')
plt.close()

labels = venn.get_labels([range(10), range(5, 15), range(3, 8), range(8, 17), range(10, 20)], fill=['number', 'logic'])
fig, ax = venn.venn5(labels, names=['list 1', 'list 2', 'list 3', 'list 4', 'list 5'])
fig.savefig('venn5.png', bbox_inches='tight')
plt.close()

labels = venn.get_labels([range(10), range(5, 15), range(3, 8), range(8, 17), range(10, 20), range(13, 25)], fill=['number', 'logic'])
fig, ax = venn.venn6(labels, names=['list 1', 'list 2', 'list 3', 'list 4', 'list 5', 'list 6'])
fig.savefig('venn6.png', bbox_inches='tight')
plt.close()



__VERSION__="ete2-2.2rev1026" 
# -*- coding: utf-8 -*-
# #START_LICENSE###########################################################
#
#
# This file is part of the Environment for Tree Exploration program
# (ETE).  http://ete.cgenomics.org
#  
# ETE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#  
# ETE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with ETE.  If not, see <http://www.gnu.org/licenses/>.
#
# 
#                     ABOUT THE ETE PACKAGE
#                     =====================
# 
# ETE is distributed under the GPL copyleft license (2008-2011).  
#
# If you make use of ETE in published work, please cite:
#
# Jaime Huerta-Cepas, Joaquin Dopazo and Toni Gabaldon.
# ETE: a python Environment for Tree Exploration. Jaime BMC
# Bioinformatics 2010,:24doi:10.1186/1471-2105-11-24
#
# Note that extra references to the specific methods implemented in 
# the toolkit are available in the documentation. 
# 
# More info at http://ete.cgenomics.org
#
# 
# #END_LICENSE#############################################################

try:
  basestring
except NameError:
  basestring = str

import re
import os
import base64 

__all__ = ["read_newick", "write_newick"]

# Regular expressions used for reading newick format
_ILEGAL_NEWICK_CHARS = ":;(),\[\]\t\n\r="
_NON_PRINTABLE_CHARS_RE = "[\x00-\x1f]+"

_NHX_RE = "\[&&NHX:[^\]]*\]"
_FLOAT_RE = "[+-]?\d+\.?\d*(?:[eE][-+]\d+)?"
#_FLOAT_RE = "[+-]?\d+\.?\d*"
_NAME_RE = "[^():,;\[\]]+"

DEFAULT_DIST = 1.0
DEFAULT_NAME = ''
DEFAULT_SUPPORT = 1.0


# Allowed formats. This table is used to read and write newick using
# different convenctions. You can also add your own formats in an easy way.
#
#
# FORMAT: [[LeafAttr1, LeafAttr1Type, Strict?], [LeafAttr2, LeafAttr2Type, Strict?],\
#    [InternalAttr1, InternalAttr1Type, Strict?], [InternalAttr2, InternalAttr2Type, Strict?]]
#
# Attributes are placed in the newick as follows:
#
# .... ,LeafAttr1:LeafAttr2)InternalAttr1:InternalAttr2 ...
#
#
#           /-A
# -NoName--|
#          |          /-B
#           \C-------|
#                    |          /-D
#                     \E-------|
#                               \-G
#
# Format 0 = (A:0.350596,(B:0.728431,(D:0.609498,G:0.125729)1.000000:0.642905)1.000000:0.567737);
# Format 1 = (A:0.350596,(B:0.728431,(D:0.609498,G:0.125729)E:0.642905)C:0.567737);
# Format 2 = (A:0.350596,(B:0.728431,(D:0.609498,G:0.125729)1.000000:0.642905)1.000000:0.567737);
# Format 3 = (A:0.350596,(B:0.728431,(D:0.609498,G:0.125729)E:0.642905)C:0.567737);
# Format 4 = (A:0.350596,(B:0.728431,(D:0.609498,G:0.125729)));
# Format 5 = (A:0.350596,(B:0.728431,(D:0.609498,G:0.125729):0.642905):0.567737);
# Format 6 = (A:0.350596,(B:0.728431,(D:0.609498,G:0.125729)E)C);
# Format 7 = (A,(B,(D,G)E)C);
# Format 8 = (A,(B,(D,G)));
# Format 9 = (,(,(,)));

NW_FORMAT = {
  0: [['name', str, True],  ["dist", float, True],    ['support', float, True],   ["dist", float, True]], # Flexible with support
  1: [['name', str, True],  ["dist", float, True],    ['name', str, True],      ["dist", float, True]], # Flexible with internal node names
  2: [['name', str, False], ["dist", float, False],   ['support', float, False],  ["dist", float, False]],# Strict with support values
  3: [['name', str, False], ["dist", float, False],   ['name', str, False],     ["dist", float, False]], # Strict with internal node names
  4: [['name', str, False], ["dist", float, False],   [None, None, False],        [None, None, False]],
  5: [['name', str, False], ["dist", float, False],   [None, None, False],        ["dist", float, False]],
  6: [['name', str, False], [None, None, False],      [None, None, False],        ["dist", float, False]],
  7: [['name', str, False], ["dist", float, False],   ["name", str, False],       [None, None, False]],
  8: [['name', str, False], [None, None, False],      ["name", str, False],       [None, None, False]],
  9: [['name', str, False], [None, None, False],      [None, None, False],        [None, None, False]], # Only topology with node names
  100: [[None, None, False],  [None, None, False],      [None, None, False],        [None, None, False]] # Only Topology
}


def format_node(node, node_type, format):
    if node_type == "leaf":
        container1 = NW_FORMAT[format][0][0]
        container2 = NW_FORMAT[format][1][0]
        converterFn1 = NW_FORMAT[format][0][1]
        converterFn2 = NW_FORMAT[format][1][1]
    else:
        container1 = NW_FORMAT[format][2][0]
        container2 = NW_FORMAT[format][3][0]
        converterFn1 = NW_FORMAT[format][2][1]
        converterFn2 = NW_FORMAT[format][3][1]

    if converterFn1 == str:
        try:
            FIRST_PART = re.sub("["+_ILEGAL_NEWICK_CHARS+"]", "_", \
                                  str(getattr(node, container1)))
        except (AttributeError, TypeError):
            FIRST_PART = "?"

    elif converterFn1 is None:
        FIRST_PART = ""
    else:
        try:
            #FIRST_PART =  "%0.6f" %(converterFn2(getattr(node, container1)))
            FIRST_PART =  "%g" %(converterFn2(getattr(node, container1)))
        except (ValueError, TypeError):
            FIRST_PART = "?"


    if converterFn2 == str:
        try:
            SECOND_PART = ":"+re.sub("["+_ILEGAL_NEWICK_CHARS+"]", "_", \
                                  str(getattr(node, container2)))
        except (ValueError, TypeError):
            SECOND_PART = ":?"
    elif converterFn2 is None:
        SECOND_PART = ""
    else:
        try:
            #SECOND_PART = ":%0.6f" %(converterFn2(getattr(node, container2)))
            SECOND_PART = ":%g" %(converterFn2(getattr(node, container2)))
        except (ValueError, TypeError):
            SECOND_PART = ":?"

    return "%s%s" %(FIRST_PART, SECOND_PART)

# Used to write into specific formats
def node2leafformat(node, format):
    safe_name = re.sub("["+_ILEGAL_NEWICK_CHARS+"]", "_", \
                             str(getattr(node, "name")))

    if format == 0 or format == 1 or format == 2 or format ==3:
        return "%s:%0.6f" %(safe_name, node.dist)
    elif format == 4 or format == 7:
        return ":%0.6f" %(node.dist)
    elif format == 5 or format == 6:
        return "%s" %(safe_name)

def node2internalformat(node, format):
    safe_name = re.sub("["+_ILEGAL_NEWICK_CHARS+"]", "_", \
                             str(getattr(node, "name")))
    if format == 0 or format == 1:
        return "%0.6f:%0.6f" %(node.support, node.dist)
    elif format == 2:
        return "%s:%0.6f" %(safe_name, node.dist)
    elif format == 3 or format == 4:
        return ":%0.6f" %(node.dist)
    elif format == 5:
        return "%s" %(safe_name)
    elif format == 6 or format == 7:
        return ""

class NewickError(Exception):
    """Exception class designed for NewickIO errors."""
    pass

def read_newick(newick, root_node=None, format=0):
    """ Reads a newick tree from either a string or a file, and returns
    an ETE tree structure.
    A previously existent node object can be passed as the root of the
    tree, which means that all its new children will belong to the same
    class as the root(This allows to work with custom TreeNode
    objects).
    You can also take advantage from this behaviour to concatenate
    several tree structures.
    """

    if root_node is None:
        from .tree import TreeNode   
        root_node = TreeNode()

    if isinstance(newick, basestring):
        if os.path.exists(newick):
            nw = open(newick, 'rU').read()
        else:
            nw = newick
        nw = nw.strip()
        if not nw.startswith('(') and nw.endswith(';'):
            return _read_node_data(nw, root_node, "single", format)
            
        elif not nw.startswith('(') or not nw.endswith(';'):
            raise NewickError('Unexisting tree file or Malformed newick tree structure.')
        else:
            return _read_newick_from_string(nw, root_node, format)

    else:
        raise NewickError("'newick' argument must be either a filename or a newick string.")

def _read_newick_from_string(nw, root_node, format):
    """ Reads a newick string in the New Hampshire format. """

    if nw.count('(') != nw.count(')'):
        raise NewickError('Parentheses do not match. Broken tree structure')

    # white spaces and separators are removed
    nw = re.sub("[\n\r\t]+", "", nw)

    current_parent = None

    # Ok, this is my own way of reading newick structures. I find it
    # more flexible and elegant than other docummented methods. Don't
    # know if I'm loosing much efficiency. It Starts by splitting the
    # structure using open parentheses. Each of the resulting chunks
    # represent an internal node. So for each chunk I create a new node
    # that hungs from the current parent node.  Each internal node chunk
    # may contain information about terminal nodes hanging from the
    # internal and clossing parenthessis (closing previously opened
    # internal nodes).
    #
    # Enjoy.
    # by JHC ;)

    # Skip the first chunk. It is always == ''
    for internal_node in nw.split("(")[1:]:
        # If this is the root of tree, use the root_node instead of
        # creating it, otherwise make a new one.
        if current_parent is None:
            current_parent = root_node
        else:
            current_parent = current_parent.add_child()
        # We can only find leaf nodes within this chunk, since rest of
        # internal nodes will be in the next newick chunks
        possible_leaves = internal_node.split(",")
        for i, leaf in enumerate(possible_leaves):
            # Any resulting sub-chunk resulting from splitting by commas can
            # be considered (tpologically) as a child to the current parent
            # node. We only discard chunks if they are empty and in the last
            # possition, meaining that the next brother is not terminal bu
            # internal node (will be visited in the next newick chunk)
            if leaf.strip() == '' and i == len(possible_leaves)-1:
                continue
            # Leaf text strings may end with a variable number of clossing
            # parenthesis. For each ')' we read the information of the
            # current node, close it and go up one more node.
            clossing_nodes = leaf.split(")")
            # first par contain leaf info
            _read_node_data(clossing_nodes[0], current_parent, "leaf", format)
            # The next parts containg clossing nodes and info about the
            # internal nodes.
            if len(clossing_nodes)>1:
                for closing_internal in clossing_nodes[1:]:
                    if closing_internal.strip() ==";": continue
                    _read_node_data(closing_internal, current_parent, "internal", format)
                    current_parent = current_parent.up
    return root_node

def _parse_extra_features(node, NHX_string):
    """ Reads node's extra data form its NHX string. NHX uses this
    format:  [&&NHX:prop1=value1:prop2=value2] """
    NHX_string = NHX_string.replace("[&&NHX:", "")
    NHX_string = NHX_string.replace("]", "")
    for field in NHX_string.split(":"):
        try:
            pname, pvalue = field.split("=")
        except ValueError as e:
            print(NHX_string, field.split("="))
            raise ValueError(e)
        node.add_feature(pname, pvalue)

def _read_node_data(subnw, current_node, node_type, format):
    """ Reads a leaf node from a subpart of the original newick
    tree """

    if node_type == "leaf" or node_type == "single":
        if node_type == "leaf":
            node = current_node.add_child()
        else:
            node = current_node
        container1 = NW_FORMAT[format][0][0]
        container2 = NW_FORMAT[format][1][0]
        converterFn1 = NW_FORMAT[format][0][1]
        converterFn2 = NW_FORMAT[format][1][1]
        flexible1 = NW_FORMAT[format][0][2]
        flexible2 = NW_FORMAT[format][1][2]
    else:
        node = current_node
        container1 = NW_FORMAT[format][2][0]
        container2 = NW_FORMAT[format][3][0]
        converterFn1 = NW_FORMAT[format][2][1]
        converterFn2 = NW_FORMAT[format][3][1]
        flexible1 = NW_FORMAT[format][2][2]
        flexible2 = NW_FORMAT[format][3][2]

    if converterFn1 == str:
        FIRST_MATCH = "("+_NAME_RE+")"
    elif converterFn1 == float:
        FIRST_MATCH = "("+_FLOAT_RE+")"
    elif converterFn1 is None:
        FIRST_MATCH = '()'

    if converterFn2 == str:
        SECOND_MATCH = "(:"+_NAME_RE+")"
    elif converterFn2 == float:
        SECOND_MATCH = "(:"+_FLOAT_RE+")"
    elif converterFn2 is None:
        SECOND_MATCH = '()'

    if flexible1:
        FIRST_MATCH += "?"
    if flexible2:
        SECOND_MATCH += "?"

    MATCH = '%s\s*%s\s*(%s)?' % (FIRST_MATCH, SECOND_MATCH, _NHX_RE)
    data = re.match(MATCH, subnw)
    if data:
        data = data.groups()
        if data[0] is not None and data[0] != '':
            node.add_feature(container1, converterFn1(data[0].strip()))

        if data[1] is not None and data[1] != '':
            node.add_feature(container2, converterFn2(data[1][1:].strip()))

        if data[2] is not None \
                and data[2].startswith("[&&NHX"):
            _parse_extra_features(node, data[2])
    else:
        raise NewickError("Unexpected leaf node format:\n\t"+ subnw[0:50] + "[%s]" %format)
    return

# def write_newick_recursive(node, features=None, format=1, _is_root=True):
#     """ Recursively reads a tree structure and returns its NHX
#     representation. """
#     newick = ""
#     if not node.children:
#         safe_name = re.sub("["+_ILEGAL_NEWICK_CHARS+"]", "_", \
#                                str(getattr(node, "name")))

#         newick += format_node(node, "leaf", format)
#         newick += _get_features_string(node, features)
#         #return newick
        
#     else:
#         if node.children:
#             newick+= "("
#         for cnode in node.children:
#             newick += write_newick(cnode, features, format=format,\
#                                      _is_root = False)
#             # After last child is processed, add closing string
#             if cnode == node.children[-1]:
#                 newick += ")"
#                 if node.up is not None:
#                     newick += format_node(node, "internal", format)
#                 newick += _get_features_string(node, features)
#             else:
#                 newick += ','
                
#     if _is_root:
#         newick += ";"
#     return newick

def write_newick(rootnode, features=None, format=1, format_root_node=True,
                 is_leaf_fn=None):
    """ Iteratively export a tree structure and returns its NHX
    representation. """
    newick = []
    leaf = is_leaf_fn if is_leaf_fn else lambda n: not bool(n.children)
    for postorder, node in rootnode.iter_prepostorder(is_leaf_fn=is_leaf_fn):
        if postorder:
            newick.append(")")
            if node.up is not None or format_root_node:
                newick.append(format_node(node, "internal", format))
                newick.append(_get_features_string(node, features))
        else:
            if node is not rootnode and node != node.up.children[0]:
                newick.append(",")
                
            if leaf(node):
                safe_name = re.sub("["+_ILEGAL_NEWICK_CHARS+"]", "_", \
                               str(getattr(node, "name")))
                newick.append(format_node(node, "leaf", format))
                newick.append(_get_features_string(node, features))
            else:
                newick.append("(")

    newick.append(";")
    return ''.join(newick)
    
def _get_features_string(self, features=None):
    """ Generates the extended newick string NHX with extra data about
    a node. """
    string = ""
    if features is None:
        features = []
    elif features == []:
        features = self.features

    for pr in features:
        if hasattr(self, pr):
            value = re.sub("["+_ILEGAL_NEWICK_CHARS+"]", "_", \
                             str(getattr(self, pr)))
            if string != "":
                string +=":"
            string +="%s=%s"  %(pr, str(value))
    if string != "":
        string = "[&&NHX:"+string+"]"

    return string


# coding utf8
import setuptools
from toolbiox.versions import get_versions

setuptools.setup(
    name="ToolBiox",
    version=get_versions(),
    author="Yuxing Xu",
    author_email="xuyuxing@mail.kib.ac.cn",
    description="a biological toolkit for genome assembly, annotation and analysis that we have accumulated from our bioinformatics work",
    url="https://github.com/SouthernCD/ToolBiox",

    packages=setuptools.find_packages(),

    install_requires=[
        "bcbio-gff>=0.6.6",
        "biopython>=1.76",
        "deepTools>=3.5.0",
        "interlap>=0.2.6",
        "matplotlib>=3.5.0",
        "networkx>=2.4",
        "numpy>=1.18.1",
        "pandas>=1.0.1",
        "pyfaidx>=0.5.5.2",
        "pysam>=0.15.2",
        "retry>=0.9.2",
        "scikit-bio>=0.5.6",
        "scikit-learn>=0.22.2.post1",
        "scipy>=1.4.1",
        "seaborn>=0.11.2",
    ],

    python_requires='>=3.5',

)



import re
import csv
import gzip
import pandas as pd
import numpy as np
from collections import OrderedDict

csv.field_size_limit(100000000)


def tsv_file_parse(file_name, key_col=0, fields="all", seq="\t", prefix="ID_", ignore_prefix=r'^#'):
    """
    This func can parse tsv-like file to dict.
    :param file_name:   path for file
    :param key_col:     which column should be key for output dict
    :param fields:      which column should be value for output dict, field can be "all" or given
                        column like "1,2,3" or "3,4,1"
    :param seq:         separator for tsv-like file
    :return: a dict with given key and value, value is a list for column chosen.
    """
    dict_output = OrderedDict()
    num = 0

    with open(file_name, newline='') as csvfile:
        spamreader = csv.reader(csvfile, delimiter=seq, quotechar='"')
        for info in spamreader:
            if len(info) == 0 or re.match(ignore_prefix, info[0]):
                continue
            if fields == "all":
                record = info
            else:
                field_list = fields.split(",")
                record = []
                for field in field_list:
                    record.append(info[int(field) - 1])

            record = tuple(record)

            if key_col == 0:
                dict_output[prefix + str(num)] = record
            else:
                dict_output[info[key_col - 1]] = record
            num = num + 1

    return dict_output


def tsv_file_parse_big(file_name, key_col=None, fields="all", seq="\t", prefix="ID_", gzip_flag=False,
                       fieldnames=None, ignore_prefix=r'^#'):
    """
    This func can parse tsv-like file to dict.
    :param file_name:   path for file
    :param key_col:     which column should be key for output dict
    :param fields:      which column should be value for output dict, field can be "all" or given
                        column like list("A","B","C")
    :param seq:         separator for tsv-like file
    :param prefix:      When key_col is None, we will get key as prefix write
    :param gzip_flag:   If this is a .gz file
    :param fieldnames:  if tsv file don't have title line, you should give it by list
    :return: a dict with given key and value, value is a list for column chosen.
    """
    # dict_output = OrderedDict()
    num = 0

    if gzip_flag is True:
        csvfile = gzip.open(file_name, 'rt')
    else:
        csvfile = open(file_name, 'r', newline='')

    spamreader = csv.DictReader(
        csvfile, delimiter=seq, quotechar='"', fieldnames=fieldnames)
    for info in spamreader:
        if len(info) == 0 or re.match(ignore_prefix, info[fieldnames[0]]):
            continue
        if fields == "all":
            record = info
        else:
            field_list = fields
            record = {}
            for i in field_list:
                record[i] = info[i]

        if key_col is None:
            ID_tmp = prefix + str(num)
        else:
            ID_tmp = info[key_col]
        num = num + 1

        # dict_output[ID_tmp] = OrderedDict.fromkeys(record)
        output_dir = OrderedDict.fromkeys(record)
        for keys in record:
            values = record[keys]
            output_dir[keys] = values
            # dict_output[ID_tmp][keys] = values

        yield output_dir

    csvfile.close()

    # return dict_output


def tsv_file_dict_parse(file_name, key_col=None, fields="all", seq="\t", prefix="ID_", gzip_flag=False,
                        fieldnames=None, ignore_head_num=0, ignore_prefix=r'^#'):
    """
    This func can parse tsv-like file to dict.
    :param file_name:   path for file
    :param key_col:     which column should be key for output dict
    :param fields:      which column should be value for output dict, field can be "all" or given
                        column like list("A","B","C")
    :param seq:         separator for tsv-like file
    :param prefix:      When key_col is None, we will get key as prefix write
    :param gzip_flag:   If this is a .gz file
    :param fieldnames:  if tsv file don't have title line, you should give it by list
    :return: a dict with given key and value, value is a list for column chosen.
    """
    dict_output = OrderedDict()
    num = 0

    if gzip_flag is True:
        csvfile = gzip.open(file_name, 'rt')
    else:
        csvfile = open(file_name, 'r', newline='')

    for i in range(0, ignore_head_num):
        next(csvfile)

    spamreader = csv.DictReader(
        csvfile, delimiter=seq, quotechar='"', fieldnames=fieldnames)
    for info in spamreader:
        if len(info) == 0 or re.match(ignore_prefix, info[spamreader.fieldnames[0]]):
            continue
        if fields == "all":
            record = info
        else:
            field_list = fields
            record = {}
            for i in field_list:
                record[i] = info[i]

        if key_col is None:
            ID_tmp = prefix + str(num)
        else:
            ID_tmp = info[key_col]
        num = num + 1

        dict_output[ID_tmp] = OrderedDict.fromkeys(record)
        for keys in record:
            values = record[keys]
            dict_output[ID_tmp][keys] = values

    csvfile.close()

    return dict_output


def tsv_file_dict_parse_big(file_name, key_col=None, fields="all", seq="\t", prefix="ID_", gzip_flag=False,
                            fieldnames=None, ignore_head_num=0, ignore_prefix=r'^#'):
    """
    This func can parse tsv-like file to dict.
    :param file_name:   path for file
    :param key_col:     which column should be key for output dict
    :param fields:      which column should be value for output dict, field can be "all" or given
                        column like list("A","B","C")
    :param seq:         separator for tsv-like file
    :param prefix:      When key_col is None, we will get key as prefix write
    :param gzip_flag:   If this is a .gz file
    :param fieldnames:  if tsv file don't have title line, you should give it by list
    :return: a dict with given key and value, value is a list for column chosen.
    """
    num = 0

    if gzip_flag is True:
        csvfile = gzip.open(file_name, 'rt')
    else:
        csvfile = open(file_name, 'r', newline='')

    for i in range(0, ignore_head_num):
        next(csvfile)

    spamreader = csv.DictReader(
        csvfile, delimiter=seq, quotechar='"', fieldnames=fieldnames)
    for info in spamreader:
        if len(info) == 0 or re.match(ignore_prefix, info[spamreader.fieldnames[0]]):
            continue
        if fields == "all":
            record = info
        else:
            field_list = fields
            record = {}
            for i in field_list:
                record[i] = info[i]

        if key_col is None:
            ID_tmp = prefix + str(num)
        else:
            ID_tmp = info[key_col]
        num = num + 1

        tmp_dict = OrderedDict.fromkeys(record)
        for keys in record:
            values = record[keys]
            tmp_dict[keys] = values

        yield (ID_tmp, tmp_dict)

    csvfile.close()


def read_list_file(file_name, ignore_prefix=r'^#', ignore_head=0):
    output_list = []
    with open(file_name, "r") as f:
        num = 0
        for each_line in f:
            num += 1
            if re.match(ignore_prefix, each_line):
                continue
            each_line = re.sub(r'\n', '', each_line)
            if each_line.strip() == '':
                continue
            if num <= ignore_head:
                continue
            output_list.append(each_line)
    return output_list


def excel_file_parse(file_name, key_col=None, prefix="ID_", fields="all", str_flag=True):
    df = pd.read_excel(file_name)
    if str_flag is True:
        df = df.astype(str)
    colname = list(df.columns)
    spamreader = df.values
    dict_output = {}
    num = 0
    for info in spamreader:
        if fields == "all":
            field_list = colname
        else:
            field_list = fields
        record = {}
        for i in field_list:
            record[i] = info[colname.index(i)]

        if key_col is None:
            ID_tmp = prefix + str(num)
        else:
            ID_tmp = info[key_col]
        num = num + 1

        dict_output[ID_tmp] = {}
        for keys in record:
            values = record[keys]
            dict_output[ID_tmp][keys] = values
    return dict_output, colname


def read_matrix_file(matrix_file, seq="\t"):
    tsv_info = tsv_file_dict_parse(matrix_file, seq=seq)

    row_key = list(tsv_info['ID_0'].keys())[0]
    col_list = list(tsv_info['ID_0'].keys())[1:]
    row_list = []

    matrix = []
    for i in tsv_info:
        data = tsv_info[i]
        row_list.append(data[row_key])
        r = []
        for c in col_list:
            r.append(float(data[c]))
        matrix.append(r)

    return np.array(matrix), col_list, row_list


def write_matrix_file(matrix, col_list, row_list, output_file):
    with open(output_file, 'w') as f:
        f.write("\t"+"\t".join(col_list)+"\n")
        for i in range(len(row_list)):
            g_id = row_list[i]
            f.write(g_id + "\t" + "\t".join([str(j)
                                             for j in matrix[i]]) + "\n")



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 27 14:15:17 2017
@author: david
"""  

import os
import glob
import numpy as np
import itertools
import argparse
import multiprocessing as mp
try: 
    import queue
except ImportError:
    import Queue as queue

from . import tree as tree_lib
    
def DetachAndCleanup(top, n):
    """
    top - the node that all the nodes in the current analysis will be below
    nUp - the node above n originally, from here to nTop needs the cached data updated
    """
    nUp = n.up
    n = n.detach()
    ch = nUp.get_children()
    if len(ch) == 1:
        if nUp.is_root():
            # can't delete the root, need to detach the child nodes and reattach them to the root
            for c in ch[0].get_children():
                nUp.add_child(c.detach())
            ch[0].delete()
            return n, nUp, top
        else:
            if top == nUp: top = nUp.up 
            nUpNew = nUp.up
            nUp.delete()
            return n, nUpNew, top
    return n, nUp, top

def GraftAndUpdate(top, n, s):
    """
    Remove n from its location in the tree and make it sister to s. Update only the sp_down data. NOT the sp_up data.
    Args:
        nTop - node being reconcilled
        n - node to move
        s - node to make n sister to.
    Returns:
        The root of the tree
    Implementation:
        n and s could be anywhere in relation to one another, need to change all the species sets that could be affected
    """
    n, nUp, top = DetachAndCleanup(top, n)
    parent = s.up
    if parent == None:
        new = tree_lib.TreeNode()
        new.add_feature("sp_up", set())
        parent = new
        top = new
        new.add_child(n)
        s = s.detach()
        new.add_child(s)
        s.sp_up = n.sp_down
    else:
        new = parent.add_child()
        new.add_child(n)
        s = s.detach()
        new.add_child(s)
    
    # sort out distances, new node will have a default distance of 1.0, correct this
    new.dist = 0.1*s.dist
    s.dist   = 0.9*s.dist
    
    # sort out sp_down data - can do a more efficient routine later if necessary (updating only what needs updating). Otherwise:
    # all nodes on path from nTop to s need sp_down updating
    new.add_feature("sp_down", s.sp_down.union(n.sp_down))
    parent.sp_down = parent.sp_down.union(n.sp_down)
    if not parent == top:
        r = parent.up
        while r != top:
            r.sp_down = set.union(*[ch.sp_down for ch in r.get_children()])
            r = r.up
    # all nodes on path from nTop to n need sp_down updating
    r = nUp
    if r != None:
        while r != top:
            r.sp_down = set.union(*[ch.sp_down for ch in r.get_children()])
            r = r.up
    return new.get_tree_root()
    
def GraftTripartAndUpdate(nTop, s1, s2, p):
    """
    Make s1 and s2 sister clades and hang the from parent node p
    Args:
        s1, s2 - nodes to be made sisters of each other
        p - node to hang (s1,s2) from
    """
    d1 = s1.up.dist
    s1, s1Parent , nTop = DetachAndCleanup(nTop, s1)   
    d2 = s2.up.dist
    s2, s2Parent, nTop = DetachAndCleanup(nTop, s2)   
    new = p.add_child()
    new.dist = 0.5*(d1+d2)
    new.add_child(s1)
    new.add_child(s2)
    new.add_feature("sp_down", s1.sp_down.union(s2.sp_down))
    n = s1Parent
    while n != nTop:
        n.sp_down = set.union(*[ch.sp_down for ch in n.get_children()])
        n = n.up
    n = s2Parent
    while n != nTop:
        n.sp_down = set.union(*[ch.sp_down for ch in n.get_children()])
        n = n.up
    nTop.sp_down = set.union(*[ch.sp_down for ch in nTop.get_children()])
    return new.get_tree_root()
        
def ContainsMonophyletic(e, O, i, I):
    """
    Check if set of species O is found only in a clade below e. If so, returns (True, n) where n is the clade otherwise (False, None)
    Args:
        e - the node to check
        O - the species set to look for
        i - the current iteration 
        I - the one-past-the-last iteration. Number of checks will be M-m
    Returns:
        qSuccess, node, depth
        where:
            qSuccess - True if found species set O to be monophyletic
            node - node of the monophyletic clade if qSuccess else the node it made it down to
            depth - depth of the node, 0=(e is O), 1=(it's child is O), etc.
    """
    E = e.sp_down
    if E == O:
        return True, e, i     
    i+=1
    if i == I or e.is_leaf(): return False, e, i-1
    ch = e.get_children()
    inChild = [bool(O & c.sp_down) for c in ch]
    if sum(inChild) > 1: return False, e, i-1         # bipart : cheange to if sum(inChild) == len(ch): return False, None 
    return ContainsMonophyletic(ch[inChild.index(True)], O, i, I)  # bipart need to check more than one potentially

def check_monophyly(node, taxa):
    """ This should be used as a wrapper of the ete method since that method returns false for a single node. It should return true
    Args:
        node - the node under which to search
        taxa - the list of taxa
    Returns
        bool - are the taxa monophyletic
    """
    if len(taxa) == 1:
        return (list(taxa)[0] in node)
    else: 
        return node.check_monophyly(target_attr='name', values=taxa)[0]

t = True
f = False

# dA = 1, dB = 0
# Subcases (S&O, S&U)
c10_a = {(f,f), (t,f), (t,t)}      
c10_b = {(f,t),}

# dA = 1, dB = 1
# Subcases (S&O, S&W, S&U)
c11_a = {(f,f,f), (t,t,t), (t,t,f), (t,f,t)}
c11_b = {(f,t,f), (f,f,t)}
c11_c = {(f,t,t),}
 
# dA = 2, dB = 0
# Subcases (S&V, S&Y, S&O, V&Y)
#c20_a = {(f,f,f,f),}
#c20_b = {(f,f,f,t), (f,f,t,f), (f,f,t,t), (f,t,f,t), (f,t,t,t), (t,f,t,t), (t,t,f,f), (t,t,f,t), (t,t,t,t)}
c20_a = {(f,f,f,f),(f,f,t,f),}
c20_b = {(f,f,f,t), (f,f,t,t), (f,t,f,t), (f,t,t,t), (t,f,t,t), (t,t,f,f), (t,t,f,t), (t,t,t,t)}
c20_c = {(f,t,f,f), (f,t,t,f), (t,f,f,f), (t,f,f,t), (t,f,t,f)}

def resolve(n, M):
    """This is an improved but also more staightforward implementation of resolve. It also considers far more sub-cases.
    It comes from writing up and generalising the first version.  
    
    Implementation:
        lower case letters are nodes, uppercase letters are species sets
    """
    if "recon" in n.features:
        sis = n.recon
        n.del_feature('recon')
        # cannot assume that sis[0] or sis[1] this will be just a single node 
        if check_monophyly(n, sis[0]) and check_monophyly(n, sis[1]):
            s0 = n.get_common_ancestor(sis[0]) if len(sis[0]) > 1 else (n&sis[0][0])
            s1 = n.get_common_ancestor(sis[1]) if len(sis[1]) > 1 else (n&sis[1][0])
            return GraftAndUpdate(n, s0, s1)
    elif "recon_2" in n.features:
        """ (dest, target, move_out)
        move target to be sister to destination, move move_out to where target was
        Note, move_out is a list of lists, corresponding to potentially multiple nodes"""
        moves = n.recon_2
        n.del_feature('recon_2')
        dest, target, move_out = moves
        if check_monophyly(n, dest) and check_monophyly(n, target) and all([check_monophyly(n, m) for m in move_out]):
            # make all the move_out hang from a single node
            if len(move_out) == 1:
                sm_new = n.get_common_ancestor(move_out[0]) if len(move_out[0]) > 1 else (n&move_out[0][0])
            else:
                sm = n.get_common_ancestor(move_out[0]) if len(move_out[0]) > 1 else (n&move_out[0][0])
                sm_new = sm.up.add_child()
                dist = sm.up.dist
                sm_new.dist = dist
                sm_new_sp_down = set()
                for m in move_out:
                    sm = n.get_common_ancestor(m) if len(m) > 1 else (n&m[0])
                    sm_new_sp_down.update(sm.sp_down)
                    sm = sm.detach()
                    sm_new.add_child(sm)
                sm_new.add_feature("sp_down", sm_new_sp_down)
            s0 = n.get_common_ancestor(dest) if len(dest) > 1 else (n&dest[0])
            s1 = n.get_common_ancestor(target) if len(target) > 1 else (n&target[0])
            d = s1.get_sisters()[0].dist
            tree = GraftAndUpdate(n, s1, s0)
            tree.dist = d
            s3 = tree.get_common_ancestor(dest + target)
            sister = s3.up.up
            top = sister.up if sister.up != None else sister
            return GraftAndUpdate(top, sm_new, sister)
        
    ch = n.get_children()
    if len(ch) != 2: return n.get_tree_root()    #bipart
    X,Y = [c.sp_down for c in ch]
    O = X&Y
    if not O: return n.get_tree_root()# no overlap, nothing to do
    if X == Y: return n.get_tree_root()  # identical species sets, they're paralogues
    sis = n.get_sisters()
    s = sis[0] if len(sis) == 1 else None
    S = set() if (n.is_root() or len(n.get_sisters()) > 1) else set.union(*[nother.sp_down for nother in n.up.get_children() if nother != n]) # works for non-binary too
    
    successA, nA, dA = ContainsMonophyletic(ch[0], O, 0, 3)   
    successB, nB, dB = ContainsMonophyletic(ch[1], O, 0, 3)   
    if dA+dB > 2:
        if dA == 2:
            successA, nA, dA = ContainsMonophyletic(ch[0], O, 0, 2)   
        if dB == 2:
            successB, nB, dB = ContainsMonophyletic(ch[1], O, 0, 2)                  
        
    if (dA==0 and dB==1) or (dA==1 and dB==0): 
        # Case A
        # a is the one that branches
        # v is the child of A that overlaps with O
        if dA == 1:
            a, b = ch
            v = nA
        else:
            b, a = ch
            v = nB
        U = set.union(*[c.sp_down for c in a.get_children() if c != v]) # allows for non-binary
        case = (bool(S&O), bool(S&U))  
        if case in c10_a:
            return GraftAndUpdate(n, b, v)
        elif case in c10_b:
            n.up.add_feature("recon", (b.get_leaf_names(),s.get_leaf_names())) 
            return n.get_tree_root()
        return n.get_tree_root()
    elif (dA==1 and dB==1): 
        # Case B
        # v is the child of A that overlaps with O (ws are the others)
        # x is the child of B that overlaps with O
        v = nA
        x = nB
        cha = ch[0].get_children()
        chb = ch[1].get_children()
        U = set.union(*[c.sp_down for c in cha if c != v]) # allows for non-binary
        W = set.union(*[c.sp_down for c in chb if c != x]) # allows for non-binary
        case = (bool(S&O), bool(S&W), bool(S&U))
        if case in c11_a:
            return GraftAndUpdate(n, x, v)
        elif case in c11_b:
            if W&S:
                # switch, so that ws, x are below a
                ws = [c for c in cha if c != v]
                w_leaves = [c.get_leaf_names() for c in ws]   # need to get the list of leaves that aren't in the node where the overlap is (for w.get_leaf_names(), below) AND need to update recon2 so that it will deal with multiple nodes in the ones it has to move
                x = v   
            else:
                ws = [c for c in chb if c != x]
                w_leaves = [c.get_leaf_names() for c in ws]
            n.up.add_feature("recon_2", (x.get_leaf_names(),s.get_leaf_names(),w_leaves)) 
            return n.get_tree_root()
        elif case in c11_c:
            ws = [c for c in chb if c != x]
            w_leaves = [g for c in ws for g in c.get_leaf_names()]
            n.up.add_feature("recon", (s.get_leaf_names(), w_leaves))  
            return n.get_tree_root()
        return n.get_tree_root()
    elif (dA==2 and dB==0) or (dA==0 and dB==2): 
        # Case C
        if dA == 2:
            a,b = ch
        else:
            b,a = ch
        cha = a.get_children()
        i = next(ii for ii,c in enumerate(cha) if (c.sp_down & O))  # there will be only one node with an overlap
        u = cha[i]
        vs = [c for ii, c in enumerate(cha) if ii != i]
        V = set.union(*[c.sp_down for c in vs])
        chu = u.get_children()  
        i = next(ii for ii,c in enumerate(chu) if (c.sp_down & O))
        z = chu[i]
        ys = [c for ii, c in enumerate(chu) if ii != i]
        Y = set.union(*[c.sp_down for c in ys])
        case = (bool(S&V), bool(S&Y), bool(S&O), bool(V&Y))
        if case in c20_a:
            return GraftAndUpdate(n, b, z)
        elif case in c20_b:
            # need to make v a single node
            if len(vs) == 1:
                vs_new = vs[0]
            else:
                vs_up = vs[0].up
                vs_new = vs_up.add_child()
                vs_new.dist = vs_up.dist
                vs_new_sp_down = set()
                for v in vs:
                    vs_new_sp_down.update(v.sp_down)
                    v = v.detach()
                    vs_new.add_child(v)
                vs_new.add_feature("sp_down", vs_new_sp_down)        
            return GraftAndUpdate(n, vs_new, b)
        elif case in c20_c:
            n.up.add_feature("recon", (b.get_leaf_names(), s.get_leaf_names()))     
            return n.get_tree_root()
        return n.get_tree_root()
    return n.get_tree_root()

def SpeciesOverlapDuplications(tree, GeneToSpecies):
    species = list(set(map(GeneToSpecies, tree.get_leaf_names())))
    genes = tree.get_leaf_names()
    gDict = {gene:i for i,gene in enumerate(genes)}
    sDict = {sp:i for i,sp in enumerate(species)}
    for n in tree.traverse('postorder'):
        if n.is_leaf(): continue
        ch = n.get_children()
        if len(ch) == 2:
            l0 = ch[0].get_leaf_names()
            l1 = ch[1].get_leaf_names()
            s0 = {GeneToSpecies(l) for l in l0}
            s1 = {GeneToSpecies(l) for l in l1}
            if s0&s1:
                n.name = "D"
            else:
                n.name = "S"
        elif len(ch) > 2:
            sp = [{GeneToSpecies(l) for l in c.get_leaf_names()} for c in ch]
            if any(sp0&sp1 for sp0, sp1 in itertools.combinations(sp, 2)):
                n.name = "D"
            else:
                n.name = "S"

def NumberOfOrthologues(tree, GeneToSpecies):
    species = list(set(map(GeneToSpecies, tree.get_leaf_names())))
    genes = tree.get_leaf_names()
    gDict = {gene:i for i,gene in enumerate(genes)}
    sDict = {sp:i for i,sp in enumerate(species)}
    orthologues = np.zeros((len(genes), len(species)))
    nOrtho = 0
    for n in tree.traverse('postorder'):
        if n.is_leaf(): continue
        ch = n.get_children()        
        if len(ch) == 2: 
            l0 = ch[0].get_leaf_names()
            l1 = ch[1].get_leaf_names()
            s0 = {GeneToSpecies(l) for l in l0}
            s1 = {GeneToSpecies(l) for l in l1}
            if s0&s1: continue
            nOrtho += len(l0)*len(l1)
            for g0 in l0:
                for s in s1:
                    orthologues[gDict[g0], sDict[s]] = 1
            for g1 in l1:
                for s in s0:
                    orthologues[gDict[g1], sDict[s]] = 1
        elif len(ch) > 2:
            for ch0, ch1 in itertools.combinations(ch, 2):
                l0 = ch0.get_leaf_names()
                l1 = ch1.get_leaf_names()
                s0 = {GeneToSpecies(l) for l in l0}
                s1 = {GeneToSpecies(l) for l in l1}
                if s0&s1: continue
                nOrtho += len(l0)*len(l1)
                for g0 in l0:
                    for s in s1:
                        orthologues[gDict[g0], sDict[s]] = 1
                for g1 in l1:
                    for s in s0:
                        orthologues[gDict[g1], sDict[s]] = 1
                
#    print(nOrtho)
#    N = (len(species)-1)*len(genes)
#    print((sum(sum(orthologues)),float(N)))
#    print(sum(sum(orthologues)))



def DoTrees(trees_queue, GeneToSpecies, out_dir, species_tree_rooted, qResolve):
    # Root using species tree if provided, otherwise tree should have been rooted already
    while True:
        try:
            trees_fn = trees_queue.get(True, 0.1)
            # print("Start: " + trees_fn)
            tree = tree_lib.Tree(trees_fn)
        #        tree = tree_lib.Tree(trees_fn, format=3)
            if len(tree) == 1: continue
            if species_tree_rooted != None:
                tree.prune(tree.get_leaf_names())
                roots = om1.GetRoots(tree, species_tree_rooted, GeneToSpecies)
                if len(roots) == 0: continue
                # Pick the first root for now
                root = roots[0]
                if root != tree:
                    tree.set_outgroup(root)
                tree.write(outfile=(out_dir + "/" + os.path.basename(trees_fn) + "_rooted.tre"))
            if qResolve:
                om1.StoreSpeciesSets(tree, GeneToSpecies)
                # Perform full reconciliation
                for n in tree.traverse("postorder"):
                    tree = resolve(n, GeneToSpecies)
            # NumberOfOrthologues(tree, GeneToSpecies)
            SpeciesOverlapDuplications(tree, GeneToSpecies)
            # print("Done: " + trees_fn)
            tree.write(outfile=(out_dir + "/" + os.path.basename(trees_fn) + ".rec.tre"), format=3)
        except queue.Empty:
            return






import math
import numpy as np
from sklearn import decomposition
from sklearn.preprocessing import StandardScaler

def hard_standard(input_matrix):
    """
    standard by average 
    each column standard, use average from sorted row
    """

    rank_matrix = []
    for i in range(input_matrix.shape[1]):
        rank_matrix.append(np.argsort(np.argsort(input_matrix[:,i])))
    rank_matrix = np.array(rank_matrix)
    rank_matrix = rank_matrix.T

    sorted_matrix = []
    for i in range(input_matrix.shape[1]):
        sorted_matrix.append(np.sort(input_matrix[:,i]))
    sorted_matrix = np.array(sorted_matrix)
    sorted_matrix = sorted_matrix.T

    avg_array = []
    for i in range(len(sorted_matrix)):
        avg_array.append(np.average(sorted_matrix[i]))
    avg_array = np.array(avg_array)

    output_matrix = []
    for i in range(rank_matrix.shape[1]):
        output_matrix.append(avg_array[rank_matrix[:,i]])
    output_matrix = np.array(output_matrix)
    output_matrix = output_matrix.T    

    return output_matrix

def euler_distance(point1, point2):
    distance = 0.0
    for a, b in zip(point1, point2):
        distance += math.pow(a - b, 2)
    return math.sqrt(distance)


def pearson_correlation(array1, array2):
    x = np.vstack((array1, array2))
    r = np.corrcoef(x)
    return r[0][1]


def PCA_do(input_list, n_components, correlation=False):
    """
    do the PCA
    input a list of lists: [[a1,a2,a3],[b1,b2,b3],...]
    a or b is sample
    1,2,3 is dimension
    n_components is the nums of PC you want get
    """
    pca = decomposition.PCA(n_components=n_components)

    input_list = np.array(input_list)

    if correlation:
        input_list = StandardScaler().fit_transform(input_list)

    return pca.fit_transform(input_list), pca.explained_variance_ratio_


def PCA_figure(input_dict, ax, correlation=False):
    """
    input_dict = {
        group_1 : [[a1,a2,a3],...],
        group_2 : [[b1,b2,b3],...]
    }
    """
    data_tran = []
    for i in input_dict:
        for j in input_dict[i]:
            data_tran.append((i, j))

    pca_input = [data for type, data in data_tran]

    pca_output, pca_ratio = PCA_do(pca_input, 2, correlation)

    output_tran = []
    for i in range(0, len(pca_output)):
        output_tran.append((data_tran[i][0], pca_output[i]))

    color_style = ['bo', 'go', 'ro', 'co', 'mo', 'yo', 'ko', 'wo']

    num = 0
    for type in input_dict:
        PC1 = [data[0] for type_tmp, data in output_tran if type_tmp == type]
        PC2 = [data[1] for type_tmp, data in output_tran if type_tmp == type]
        ax.plot(PC1, PC2, color_style[num], markersize=3, label=type)
        num = num + 1

    ax.legend(loc='upper left')

    ax.xlabel("PC1  %.1f%%" % (pca_ratio[0] * 100))
    ax.ylabel("PC2  %.1f%%" % (pca_ratio[1] * 100))

    # plt.savefig(figure_file, dpi=1000)

    return ax, output_tran, pca_ratio
